from anthropic import BaseModel
from autogen_core.models import UserMessage


class PublicMessage(BaseModel):
    """Message containing the user prompt and triggering the singleton agent."""
    body: UserMessage



import asyncio
import copy
import re
import uuid
from typing import List, Tuple, Optional, Union

from autogen_core import RoutedAgent, message_handler, MessageContext, DefaultTopicId, FunctionCall
from autogen_core.models import SystemMessage, LLMMessage, AssistantMessage, FunctionExecutionResultMessage, CreateResult

from spoox.agents.agent_system import AgentSystem
from spoox.agents.errors import ModelClientError, MaxOnlyTextMessagesError, MaxIterationsError, AgentError
from spoox.agents.mas.agents.prompts import get_AGENT_FAILED_GROUP_CHAT_MESSAGE
from spoox.agents.mas.messages import GroupChatMessage, RequestToSpeak, GROUP_CHAT_TOPIC_TYPE
from spoox.environment.model_clients.custom_response_types import CreateResultOpenAI, AssistantMessageOpenAI

# to ensure progress and prevent endless text-only replies, a limit on consecutive text-only messages is enforced
MAX_ONLY_TEXT_MESSAGES = 3

# if the model client encounters errors, the agent allows a maximum of three retry attempts
MAX_MODEL_CLIENT_ERRORS_RETRIALS = 3


class BaseGroupChatAgent(RoutedAgent):
    """
    Base agent class used to build agents that follow the concepts and design principles of the spoox framework.
    Agents maintain a local chat history by recording all distributed GroupChatMessages.
    When a RequestToSpeak is received, the agent starts its internal execution loop.
    """

    def __init__(
            self,
            description: str,
            system_message: str,
            agent_system: AgentSystem,
            next_agent_topic_types: list[str] = None,
            max_internal_iterations: int = 50,
            fallback_agent_topic_type: str = None,
            reset_on_request_to_speak: bool = True,
            only_track_user_messages: bool = False,
            skip: bool = False # todo remove
    ) -> None:
        """
        Base agent class used to build agents following the concepts and design principles of the spoox framework.

        Args:
            description (str): One-sentence agent description passed to AutoGen's RoutedAgent.
            system_message (str): System message added as the initial message to the agent's message history.
            agent_system (AgentSystem): Agent system associated with the agent, providing access to the environment, model client, and other shared components.
            next_agent_topic_types (list[str]): List of all possible next agent topic types that the agent is allowed to call.
            max_internal_iterations (int): Maximum number of internal iterations the agent may perform, corresponding to the maximum number of LLM calls.
            fallback_agent_topic_type (str): Topic type of the agent to be invoked if this agent fails.
            reset_on_request_to_speak (bool): If True, internal messages are cleared from the chat history each time the agent is called, while group chat messages remain.
            only_track_user_messages (bool): If True, the agent's group chat history only keeps track of user messages.
            skip (bool): If True and if the agent is called, it just returns.
        """

        super().__init__(description=description)
        self._next_agent_topic_types = [n.lower() for n in next_agent_topic_types or []]
        self._max_internal_iterations = max_internal_iterations
        self._fallback_agent_topic_type = fallback_agent_topic_type
        self._reset_on_request_to_speak = reset_on_request_to_speak
        self._only_track_user_messages = only_track_user_messages
        self._skip = skip  # todo remove

        self._environment = agent_system.environment
        self._model_client = agent_system.model_client
        self._interface = agent_system.interface
        self._usage_stats = agent_system.usage_stats
        self._save_logs_f = agent_system.save_logs
        self._return_next_time_possible_event = agent_system.timeout_event
        self._tools = self._environment.get_tools(self) if self._environment else []

        self._chat_history: List[LLMMessage] = [SystemMessage(content=system_message)]
        self._chat_history_group_chat_only: List[LLMMessage] = [SystemMessage(content=system_message)]

        # logging
        self._interface.print_logging(system_message, f"logging - {self.id.type} - system_message")
        for t in self._tools:
            self._interface.print_logging(str(t.schema), f"logging - {self.id.type} - tool_schema")

    @message_handler
    async def handle_group_chat_message(self, message: GroupChatMessage, ctx: MessageContext) -> None:
        """
        Each agent keeps track of the entire group chat in its internal message history.
        Therefore, it stores every incoming GroupChatMessage and tracks which agent posted each message.
        Thereby, `_chat_history` stores all group chat messages as well as all internal message history
        and `_chat_history_group_chat_only` only tracks GroupChatMessages.
        This ensures that when the chat history is reset (controlled by `reset_on_request_to_speak`),
        the chat history is simply replaced with `_chat_history_group_chat_only`,
        so that only group chat messages are retained and all internal iteration messages are discarded.
        """
        if self._only_track_user_messages and message.body.source.lower() != 'user':
            return
        new_messages = [
            SystemMessage(content=f"Transferred to {message.body.source.capitalize()} agent."),
            message.body,
        ]
        self._chat_history_group_chat_only.extend(new_messages)
        if message.body.source != self.id.type:
            self._chat_history.extend(new_messages)

    @message_handler
    async def handle_request_to_speak(self, message: RequestToSpeak, ctx: MessageContext) -> None:
        """
        Agent is requested to speak: parts of its internal state are reset, and the internal execution loop is started.
        Furthermore, if the agent loop throws errors, they are caught and logged, and a fallback mechanism is triggered.
        """
        if self._skip:
            return # todo remove

        # ensures the env is fully reset to prevent any influence from previous agents that used the same env
        if self._environment:
            await self._environment.reset()

        # reset chat history to group chat messages only; all previous internal iteration messages are discarded
        if self._reset_on_request_to_speak:
            self._chat_history = copy.deepcopy(self._chat_history_group_chat_only)
            self._interface.print_logging(
                "reset to group chat history only on request to speak", f"logging - {self.id.type} - reset")

        # add a system message that instructs the model to adopt this agent's persona
        self._chat_history.append(
            SystemMessage(content=f"Transferred to {self.id.type.capitalize()} agent, adopt the persona immediately.")
        )

        # additional chat history logging
        logging_chat_hist = ' -> '.join(
            f"[{cleaned[:60]}…]"
            for h in self._chat_history
            for cleaned in [str(h.content or "").replace("\n", "").replace("\\", "")]
        )
        self._interface.print_logging('[start] -> ' + logging_chat_hist, f"logging - {self.id.type} - chat history")

        # run the agent's internal loop;
        # if agent loop fails, no final group chat message is generated, and the fallback agent is called if available
        try:
            await self._agent_loop(ctx)
            return
        except AgentError as e:
            self._interface.print_highlight(str(e), "Agent Error")
            self._usage_stats["agent_errors"].append(e)
        except Exception as e:
            self._interface.print_highlight(str(e), "Unexpected Error")
            self._usage_stats["agent_errors"].append(e)

        # fallback mechanism
        if self._fallback_agent_topic_type:
            failure_message = get_AGENT_FAILED_GROUP_CHAT_MESSAGE(self.id.type, self._fallback_agent_topic_type)
            self._interface.print_shadow(failure_message, "Fallback Agent Call")
            await self._send_group_chat_message_and_request_to_speak(failure_message, self._fallback_agent_topic_type)

        # if error and no fallback -> just return -> no next agent will be triggered -> autogen runtime exits

    async def _agent_loop(self, ctx: MessageContext):
        """
        Request the LLM, process its response, and repeat until the agent has finished.
        High-level flow:

        1.	The model client (LLM) is queried and returns a response.
        2.	The response is checked for tool calls. If tool calls are present, the corresponding tools are executed and the loop restarts.
        3.	If the response consists of a plain text answer, it is checked for a requested next-agent tag and whether the referenced agent exists.
        """

        # tracking consecutive model client errors and LLM "only-text" responses
        counter_only_text_messages = 0
        model_client_errors = 0

        for i in range(1, self._max_internal_iterations + 1):

            # checking agent system timeout event
            if self._return_next_time_possible_event.is_set():
                return

            # logging
            self._save_logs_f()
            self._usage_stats['llm_calls_count'] += 1

            # request model client (llm)
            llm_res, model_client_errors = await self._request_llm(ctx, model_client_errors)
            if llm_res is None:
                continue
            self._interface.print_logging(str(llm_res), f"logging - {self.id.type} - entire llm_res")
            content = llm_res.content

            # add the response to session and print thoughts if available
            if isinstance(llm_res, CreateResultOpenAI):
                self._chat_history.append(
                    AssistantMessageOpenAI(content=content, thought=llm_res.thought,
                                           source=self.id.type, response_items=llm_res.response_items))
            else:
                self._chat_history.append(
                    AssistantMessage(content=content, thought=llm_res.thought, source=self.id.type))
            if llm_res.thought:
                self._interface.print_thought(llm_res.thought, f"{self.id.type} - thoughts")

            # check if list of tool calls
            tool_results_message = await self._exec_tools(ctx, content)
            if tool_results_message is not None:
                counter_only_text_messages = 0
                self._chat_history.append(tool_results_message)
                # trigger LLM again with tool results in _chat_history
                continue

            # check if just text (autogen: if it is not a list of tool calls, it has to be string)
            assert isinstance(content, str)
            self._interface.print(content, f"{self.id.type} - message")

            # check if agent finished and calls next agent
            next_agent_tag = self._includes_agent_tag(content)
            if next_agent_tag is not None:
                # we assume that if an agent tag is included, this message contains the summary for the group chat
                await self._send_group_chat_message_and_request_to_speak(content, next_agent_tag)
                return

            # check if no `_next_agent_topic_types` were defined; if so, the agent finishes if no tools are called
            if not self._next_agent_topic_types:
                await self._send_group_chat_message(content)
                return

            # check if MAX_ONLY_TEXT_MESSAGES is reached
            counter_only_text_messages += 1
            if counter_only_text_messages > MAX_ONLY_TEXT_MESSAGES:
                if counter_only_text_messages == 2:
                    # additional reminder after 2 only_text_messages
                    tags = ', '.join([f"[{f}]" for f in self._next_agent_topic_types])
                    self._chat_history.append(SystemMessage(
                        content=f"Remember: do not ask the user for feedback or confirmation. "
                                f"Do your work as best as possible. When you believe your task is complete, "
                                f"call the appropriate next agent using one of the predefined tags (available tags: {tags})"
                    ))
                raise MaxOnlyTextMessagesError(self.id.type, MAX_ONLY_TEXT_MESSAGES)

        raise MaxIterationsError(self.id.type, self._max_internal_iterations)

    async def _request_llm(self, ctx: MessageContext, model_client_errors: int) -> Tuple[Optional[CreateResult], int]:
        """Invokes the model client (LLM) and handles any exception that occurs."""

        try:
            llm_res = await self._model_client.create(
                messages=self._chat_history,
                tools=self._tools,
                cancellation_token=ctx.cancellation_token,
            )
        except Exception as e:
            self._usage_stats['model_client_exceptions'].append(e)
            self._interface.print_logging(str(e), "Model Client Error")
            if model_client_errors >= MAX_MODEL_CLIENT_ERRORS_RETRIALS:
                raise ModelClientError(self.id.type, MAX_MODEL_CLIENT_ERRORS_RETRIALS, e)
            else:
                return None, model_client_errors + 1
        # llm call success
        self._usage_stats['prompt_tokens'].append(llm_res.usage.prompt_tokens)
        self._usage_stats['completion_tokens'].append(llm_res.usage.completion_tokens)
        return llm_res, 0

    async def _exec_tools(self, ctx: MessageContext, content: Union[str, List[FunctionCall]]) -> Optional[FunctionExecutionResultMessage]:
        """Executes available tool calls, if any."""

        # check whether the LLM response contains tool calls
        if not isinstance(content, list) or not all(isinstance(c, FunctionCall) for c in content):
            return None
        if self._environment is None:
            return None

        # execute tool calls
        tool_results = await asyncio.gather(
            *[
                self._environment.execute_tool_call(
                    self._tools, c, ctx.cancellation_token, self._interface, self._usage_stats, self.id.type
                )
                for c in content
            ]
        )
        return FunctionExecutionResultMessage(content=tool_results)

    def _includes_agent_tag(self, message: str) -> Optional[str]:
        """Checks if the message includes an agent tag and returns the first match detected."""

        for nt in self._next_agent_topic_types:
            pattern = rf"\[[^\]]*{re.escape(nt)}[^\]]*\]"
            if re.search(pattern, message, flags=re.IGNORECASE):
                self._usage_stats['next_agent_calling_chain'].append(nt)
                return nt
        return None

    async def _send_group_chat_message(self, message: str):
        self._usage_stats['group_chat_message_lengths'].append(len(message))
        await self.publish_message(
            message=GroupChatMessage(
                nonce=str(uuid.uuid4()),
                body=AssistantMessage(content=message, source=self.id.type),
            ),
            topic_id=DefaultTopicId(type=GROUP_CHAT_TOPIC_TYPE),
        )

    async def _send_request_to_speak(self, agent_type: str):
        await self.publish_message(
            RequestToSpeak(nonce=str(uuid.uuid4())), DefaultTopicId(type=agent_type)
        )

    async def _send_group_chat_message_and_request_to_speak(self, message: str, agent_type: str):
        # remove tag from message
        pattern = rf"\[[^\]]*{re.escape(agent_type)}[^\]]*\]"
        message = re.sub(pattern, "", message)
        await self._send_group_chat_message(message)
        # 0.1 delay to ensure the GroupChatMessage can be observed before the RequestToSpeak
        # (I think it is not required, however, it certainly does not hurt)
        await asyncio.sleep(0.1)
        await self._send_request_to_speak(agent_type)



from spoox.agents.prompts import get_CODE_EXECUTION_CAPABILITIES, MORE_OPERATING_RULES


MAS_CONTEXT = [
    f"""""",
    f"""# Context""",
    f"""- You are a helpful agent part of a Multi Agent System that solves server and command-line related tasks.""",
    f"""- You, all other agents and the user operate on the same system.""",
    f"""- Once the user provides a task, the Multi Agent System's goal is to complete the task collaboratively, without returning to the user for clarification."""
    f"""- The chat history includes the user's task, progress summaries, and information from previous agents, such as prior sub-task plans, feedback, executed steps, exploration details and more.""",
]


def get_EXPLORER_SYSTEM_MESSAGE(agent_role: str, next_agent_topic_type: str, additional_tool_descriptions: [str],
                                support_feedback: bool = False) -> str:
    _TASK_DESCR_1 = [
        f"""""",
        f"""# Role and Task Description""",
        f"""- Role: you are the **{agent_role.capitalize()}** agent of the Multi Agent System.""",
        f"""- Based on the provided task by the user, use the given tools to gather some basic information relevant to completing the task.""",
    ]
    _TASK_DESCR_2 = [
        "- It is possible that the last prior agent has left instructions for you in latest message with open topics that still need to be explored. If so, prioritize answering the other agents' exploration requests."
    ]
    _TASK_DESCR_3 = [
        f"""- IMPORTANT: **Do NOT try to solve the tasks**, instead, collect only basic information that will assist later agents in planning and actively solving the task.""",
        f"""- When finished write a concise summary of only the gathered information and include the tag '[{next_agent_topic_type}]'.""",
        f"""- Include a tag in your answer only when you intend to pass it on."""
    ]
    _TASK_DESCR = _TASK_DESCR_1 + _TASK_DESCR_2 + _TASK_DESCR_3 if support_feedback else _TASK_DESCR_1 + _TASK_DESCR_3

    return '\n'.join(
        MAS_CONTEXT +
        _TASK_DESCR +
        get_CODE_EXECUTION_CAPABILITIES(additional_tool_descriptions) +
        MORE_OPERATING_RULES
    )


def get_SOLVER_SYSTEM_MESSAGE(agent_role: str, test_agent_topic_type: str, additional_tool_descriptions: [str]) -> str:
    _TASK_DESCR = [
        f"""""",
        f"""# Role and Task Description""",
        f"""- Role: you are the **{agent_role.capitalize()}** agent of the Multi Agent System.""",
        f"""- Your job is to **actively** complete the user’s task using the available tools.""",
        f"""- The chat history contains the user's task, along with any progress summaries and information gathered by previous agents.""",
        f"""- Continue working until the task is done or no further contribution is possible.""",
        f"""- When finished, or if no more progress can be made, write a summary of what you did and include the tag '[{test_agent_topic_type}]'.""",
        f"""- Include a tag in your answer only when you intend to pass it on.""",
    ]
    return '\n'.join(
        MAS_CONTEXT +
        _TASK_DESCR +
        get_CODE_EXECUTION_CAPABILITIES(additional_tool_descriptions) +
        MORE_OPERATING_RULES
    )


def get_SUB_TASK_SOLVER_SYSTEM_MESSAGE(agent_role: str, planner_agent_topic_type: str,
                                       additional_tool_descriptions: [str]) -> str:
    _TASK_DESCR = [
        f"""""",
        f"""# Role and Task Description""",
        f"""- Role: you are the **{agent_role.capitalize()}** agent of the Multi Agent System.""",
        f"""- Your job is to **actively** complete only the most recently defined sub-task, using the available tools.""",
        f"""- The previous {planner_agent_topic_type.capitalize()} agent message includes the sub-task for you to actively complete, and an optional plan sketch you **may** follow.""",
        f"""- Attention: you do not have to follow the proposed plan if you think you have a better idea  but always keep your focus on the sub-task.""",
        f"""- Continue working until the described sub-task is done or no further contribution is possible.""",
        f"""- Also, verify whether the chosen sub-task is reasonable. If it needs refinement, provide your feedback and include the tag '[{planner_agent_topic_type}]'.""",
        f"""- When finished, or you get stuck, write a summary of what you did and include the tag '[{planner_agent_topic_type}]'.""",
        f"""- Include a tag in your answer only when you intend to pass it on.""",
    ]
    return '\n'.join(
        MAS_CONTEXT +
        _TASK_DESCR +
        get_CODE_EXECUTION_CAPABILITIES(additional_tool_descriptions) +
        MORE_OPERATING_RULES
    )


def get_SUB_TASK_PLANNER_SYSTEM_MESSAGE(agent_role: str, explorer_topic_type: str, solver_agent_topic_type: str,
                                        tester_agent_topic_type: str) -> str:
    _TASK_DESCR = [
        f"""""",
        f"""# Role and Task Description""",
        f"""- Role: you are the **{agent_role.capitalize()}** agent of the Multi Agent System.""",
        f"""- Your job is to define the next reasonable and small **sub-task** of the overall user task and provide a **high-level** plan outline of how it could be solved.""",
        f"""- The {solver_agent_topic_type.capitalize()} agent will carry out the sub-task, you should **only** focus on determine the next appropriate sub-task and a simple plan sketch.""",
        f"""- Make sure the sub-task and plan sketch describe only the next reasonable step forward, keeping in mind that some work may already have been completed by previous agents.""",
        f"""- The sub-task should be described concisely in verbal form and make sure the plan outline is high-level only and **not** a detailed breakdown.""",
        f"""- If you can not provide the next sub-task and need more system or environment information first, specify exactly what you need and include the tag '[{explorer_topic_type}]'.""",
        f"""- If you want your finalized sub-task carried out, you must describe it, give a high-level plan sketch, and also include the tag '[{solver_agent_topic_type}]'.""",
        f"""- If no further sub-tasks are needed and the overall user task is complete, respond only with the tag '[{tester_agent_topic_type}]'.""",
        f"""- Include a tag in your answer only when you intend to pass it on.""",
    ]
    return '\n'.join(
        MAS_CONTEXT +
        _TASK_DESCR
    )


def get_TESTER_SYSTEM_MESSAGE(agent_role: str, previous_agent_topic_type: str, next_agent_topic_type: str,
                              additional_tool_descriptions: [str]) -> str:
    _TASK_DESCR = [
        f"""""",
        f"""# Role and Task Description""",
        f"""- Role: you are the **{agent_role.capitalize()}** agent of the Multi Agent System.""",
        f"""- The given task is already solved by the previous agents.""",
        f"""- Your task is now to test if the overall user's task was completed successfully.""",
        f"""- Use the information provided by previous agents, along with your available tools, to test the solution.""",
        f"""- If you come to the conclusion that the task is **not** solved successfully and entirely, describe concisely why it is not solved and include the tag '[{previous_agent_topic_type}]'.""",
        f"""- If the task is solved successfully write a summary of what you did and include the tag '[{next_agent_topic_type}]'.""",
        f"""- Include a tag in your answer only when you intend to pass it on.""",
    ]
    return '\n'.join(
        MAS_CONTEXT +
        _TASK_DESCR +
        get_CODE_EXECUTION_CAPABILITIES(additional_tool_descriptions) +
        MORE_OPERATING_RULES
    )


def get_REFINER_SYSTEM_MESSAGE(agent_role: str, tester_agent_topic_type: str,
                               approver_agent_topic_type: str, additional_tool_descriptions: [str]) -> str:
    _TASK_DESCR = [
        f"""""",
        f"""# Role and Task Description""",
        f"""- Role: you are the **{agent_role.capitalize()}** agent of the Multi Agent System.""",
        f"""- Your job is to resolve discovered bugs and refine and complete the already implemented solution of the user's task.""",
        f"""- The previous agent's latest message includes already identified bugs, missing parts, and other errors in the user task's solution. Concentrate solely on resolving those latest findings.""",
        f"""- If you are finished with refining and fixing, or you get stuck, write a summary of what you did and include the tag '[{tester_agent_topic_type}]'.""",
        f"""- If the latest prior agent's message reports no errors, respond only with the tag '[{approver_agent_topic_type}]'.""",
        f"""- Include a tag in your answer only when you intend to pass it on.""",
    ]
    return '\n'.join(
        MAS_CONTEXT +
        _TASK_DESCR +
        get_CODE_EXECUTION_CAPABILITIES(additional_tool_descriptions) +
        MORE_OPERATING_RULES
    )


def get_APPROVER_SYSTEM_MESSAGE(agent_role: str, solver_agent_topic_type: str, next_agent_topic_type: str,
                                additional_tool_descriptions: [str]) -> str:
    _TASK_DESCR = [
        f"""""",
        f"""# Role and Task Description""",
        f"""- Your role is the "{agent_role.capitalize()}" agent of the Multi Agent System.""",
        f"""- The task has already been completed by earlier agents. Your responsibility is now to evaluate and test the implemented solution.""",
        f"""- Ensure that all specifications are implemented precisely as described in the user's initial request.""",
        f"""- Be critical and selective, and only approve the solution if all requirements are fully and precisely satisfied.""",
        f"""- If you decide that the task was **not solved successfully**, briefly explain why and include the tag '[{solver_agent_topic_type}]'.""",
        f"""- If the entire task is solved successfully write a summary of what you did and include the tag '[{next_agent_topic_type}]'.""",
    ]
    return '\n'.join(
        MAS_CONTEXT +
        _TASK_DESCR +
        get_CODE_EXECUTION_CAPABILITIES(additional_tool_descriptions) +
        MORE_OPERATING_RULES
    )


def get_SUMMARIZER_SYSTEM_MESSAGE(agent_role: str) -> str:
    _TASK_DESCR = [
        f"""""",
        f"""# Role and Task Description""",
        f"""- Your role is the "{agent_role.capitalize()}" agent of the Multi Agent System.""",
        f"""- Your job is to write the final concise summary of the entire task completion process.""",
        f"""- Your summary should **briefly** describe the solution plan, its execution, and the results of testing.""",
        f"""- Highlight key findings or lessons learned, if any.""",
    ]
    return '\n'.join(
        MAS_CONTEXT +
        _TASK_DESCR
    )


def get_AGENT_FAILED_GROUP_CHAT_MESSAGE(agent_role: str, fallback_agent: str = None) -> str:
    _ERROR_MESSAGE = f"{agent_role.capitalize()} agent failed and could not add any contribution."
    if fallback_agent:
        return f"{_ERROR_MESSAGE} -> fallback agent called: {fallback_agent.capitalize()}"
    return _ERROR_MESSAGE



import copy
import sys
import questionary
import yaml

from yaspin import yaspin
from importlib.metadata import version, PackageNotFoundError
from pathlib import Path
from typing import Optional
from art import tprint
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel

from spoox.environment.local_environment import ConfirmationMode
from spoox.utils import AgentSystemId
from spoox.environment.model_clients.utils import ModelClientId

_console = Console()
_spinner = None

CONFIG_FORM = {
    'model_client_id': {
        'type': 'choice',
        'prompt': "Please select a model client:",
        'choices': [e.value for e in list(ModelClientId)],
        'value': None,
    },
    'model_id': {
        'type': 'input',
        'prompt': "Please provide a model id (e.g. 'claude-sonnet-4-5', 'gpt-5-mini'):",
        'value': None,
    },
    'agent_id': {
        'type': 'choice',
        'prompt': "Please select an agent system:",
        'choices': [e.value for e in list(AgentSystemId)],
        'default': AgentSystemId.SPOOX_M.value,
        'value': None,
    },
    'confirmation_mode': {
        'type': 'choice',
        'prompt': "Please select confirmation mode – [strict]: confirmation before every env interaction – [self_evaluation]: agents autonomously decide when confirmation is needed (e.g., `cd` does not require confirmation, while `rm` commands always do) – [no_confirmation]: agent never seeks user confirmation, warning: use only in isolated or sandboxed environments.",
        'choices': [e.value for e in list(ConfirmationMode)],
        'default': ConfirmationMode.SELF_EVALUATION.value,
        'value': None,
    },
    'debugging_mode': {
        'type': 'choice',
        'prompt': "Activate debugging mode:",
        'choices': ['no', 'yes'],
        'default': 'no',
        'value': None,
    }
}


def print_cli_header():
    """Print spoox cli static header."""

    try:
        spoox_version = version('spoox')
    except PackageNotFoundError:
        spoox_version = None

    _console.rule(characters="—", style="grey30")
    tprint("spoox CLI", font='tarty2', space=1)  # decent font options: 'soft', 'tarty2'
    _console.print("👻  Welcome to the spoox CLI", style="bold")
    if spoox_version is not None:
        _console.print(f"👻  Version: {spoox_version}", style="dim")
    _console.print("👻  GitHub: https://github.com/plaume8/spoox", style="dim")
    _console.print("")
    _console.print("👻  Spoox CLI is a terminal-integrated, LLM-powered multi-agent system that assists with tasks",
                   style="dim")
    _console.print("👻  ranging from simple OS operations to complex SE challenges, directly in the terminal.",
                   style="dim")
    _console.print("👻  The integrated agent systems are based on the spoox MAS design framework,", style="dim")
    _console.print("👻  a generic architectural framework for multi-agent topology and communication design.",
                   style="dim")
    _console.print("")
    _console.rule(characters="—", style="grey30")
    _console.print("")


def print_cli_footer(agent_id: str):
    """Print spoox cli static footer."""

    _console.print(f"👻  Agent system [orange1]{agent_id}[/orange1] initialized successfully.", style="dim")
    _console.print("👻  Typical use cases include:", style="dim")
    _console.print("👻  - ‘Analyze the Apache logs and answer the question …’", style="dim")
    _console.print("👻  - ‘For my newly created Python script, create a comprehensive test suite …’", style="dim")
    _console.print("👻  - ‘I configured a Node server but it continues to fail. Help me fix it …’", style="dim")
    _console.print("")
    _console.print("👻  Ready to get to work! Just type in your task, question, or challenge.", style="bold")
    _console.print(f"👻  Type [orange1]q[/orange1] exit.", style="dim")


def print_error_message(message: str):
    _console.print("")
    _console.print(f"👻  {message}", style="bold red")
    _console.print("")


def start_loading():
    """Start a loading animation in the terminal."""
    global _spinner
    _spinner = yaspin(text="Loading...", color="blue")
    _spinner.start()


def stop_loading():
    """Stop the loading animation."""
    global _spinner
    if _spinner:
        _spinner.stop()


def clear_lines(num_lines: int):
    """Clear the specified number of lines from the terminal."""
    for _ in range(num_lines):
        sys.stdout.write("\033[F")  # move cursor up
        sys.stdout.write("\033[K")  # clear line
    sys.stdout.flush()


async def confirm_cli_config(config: dict, logs_dir: Path, cache_config: bool = True) -> dict:
    """CLI for filling, confirming and caching a valid spoox config."""

    # if any value was filled already in the config, request the remaining attributes;
    # otherwise load cached config or create entirely new config
    config_cache = logs_dir / "config_cache.yaml"
    if any(f['value'] is not None for f in config.values()):
        config = await fill_cli_config(config)
    else:
        cached_config = load_cached_cli_config(config_cache)
        if cached_config is None:
            config = await fill_cli_config(config)
        else:
            config = cached_config

    # ask user to confirm config
    print_config(config, "Selected configuration:")
    if not questionary.confirm("Please confirm the config ?", qmark='👻 ').ask():
        config_cache.unlink(missing_ok=True)
        config = await confirm_cli_config(copy.deepcopy(CONFIG_FORM), logs_dir, cache_config=False)

    if cache_config:
        # cache config
        with config_cache.open("w") as file:
            yaml.dump(config, file)
        # clear config process from terminal and print summary
        _console.print("")
        _console.rule(characters="—", style="grey30")
        _console.print("")
    return config


def print_config(config: dict, title: str) -> None:
    """Print configuration in terminal."""
    _console.print(f"👻  {title}")
    md = Markdown(f"```yaml\n{'\n'.join([f"{a}: {f['value']}" for a, f in config.items()])}\n```")
    _console.print(Panel(md, style='#555555'))


def load_cached_cli_config(config_cache: Path) -> Optional[dict]:
    """Load cached config."""

    last_config = None
    if config_cache.is_file():
        try:
            with config_cache.open("r") as file:
                last_config = yaml.safe_load(file)
        except Exception:
            pass
    if last_config is not None:
        if not isinstance(last_config, dict):
            last_config = None
        elif any(a not in last_config.keys() for a in CONFIG_FORM.keys()):
            last_config = None
    if last_config is not None:
        _console.print("👻  We found a cached spoox CLI configuration.", style="dim")
    return last_config


async def fill_cli_config(config: dict, request_all: bool = True) -> dict:
    """CLI process for filling a spoox config."""

    # get all user inputs
    _console.print("👻  Complete the following steps to config the spoox CLI:", style="dim")
    for id, form in config.items():
        user_input = None
        default = form.get('default', None)
        if form['value'] is not None:
            continue
        if default is not None and not request_all:
            user_input = form['default']
        elif form['type'] == 'choice':
            user_input = questionary.select(form['prompt'], form['choices'], default=default, qmark='👻 ').ask()
        elif form['type'] == 'input':
            user_input = questionary.text(form['prompt'], default=(default or ""), qmark='👻 ').ask()
        config[id]['value'] = user_input

    # validations
    # todo check if model exists for model_id

    return config



from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.agents.prompts import get_APPROVER_SYSTEM_MESSAGE
from spoox.agents.mas.base_agent import BaseGroupChatAgent


class ApproverAgent(BaseGroupChatAgent):
    """
    The Approver's role is to examine the implemented solution in detail
    and decide whether the overall task has been completed entirely.
    He has no access to previous agent summaries, only the user's initial task description is provided.
    """

    def __init__(
            self,
            topic_type: str,
            agent_system: AgentSystem,
            solver_agent_topic_type: str,
            next_agent_topic_type: str,
    ) -> None:

        next_agent_topic_types = [solver_agent_topic_type, next_agent_topic_type]

        system_message = get_APPROVER_SYSTEM_MESSAGE(
            topic_type, solver_agent_topic_type, next_agent_topic_type,
            agent_system.environment.get_additional_tool_descriptions(self))

        super().__init__(
            description="Agent tasked to decide if the implementation is complete and correct.",
            system_message=system_message,
            agent_system=agent_system,
            next_agent_topic_types=next_agent_topic_types,
            max_internal_iterations=200,
            only_track_user_messages=True
        )



import os
from enum import Enum

from autogen_core.models import ChatCompletionClient
from autogen_ext.models.anthropic import AnthropicChatCompletionClient
from autogen_ext.models.ollama import OllamaChatCompletionClient

from spoox.environment.model_clients.custom_client_openai import CustomOpenAIResponseAPIClient


class ModelClientId(Enum):
    """All available model client ids."""

    ANTHROPIC = 'anthropic'
    OLLAMA = 'ollama'
    OPENAI = 'openai'


def setup_model_client(client_id: ModelClientId, model_id: str) -> ChatCompletionClient | CustomOpenAIResponseAPIClient:
    """
    Based on the provided client_id and model_id, the corresponding model client instance is created.

    Args:
        client_id (str): The base model client, options: 'ollama', 'openai', 'anthropic'.
        model_id (str): The actual model id (e.g. 'qwen3:8b', 'claude-sonnet-4-5-20250929').

    Returns:
        ChatCompletionClient: Model client ready to be used by the agent system.
    """

    _check_env(client_id)

    if client_id == ModelClientId.OLLAMA:
        # get ollama endpoint
        host = os.environ['OLLAMA']
        # special exception for gpt-oss models -> ollama not keeps a pre-set model_info -> todo test if still necessary
        if model_id in ["gpt-oss:20b", "gpt-oss:120b"]:
            model_info = {
                "vision": False,
                "function_calling": True,
                "json_output": True,
                "family": "unknown",
                "structured_output": True,
                "multiple_system_messages": False
            }
            return OllamaChatCompletionClient(model=model_id, model_info=model_info, host=host)
        return OllamaChatCompletionClient(model=model_id, host=host)

    if client_id == ModelClientId.OPENAI:
        return CustomOpenAIResponseAPIClient(model_id=model_id)

    if client_id == ModelClientId.ANTHROPIC:
        return AnthropicChatCompletionClient(model=model_id)

    raise ValueError(f"No model client could be set up for: '{client_id}', '{model_id}'.")


def _check_env(client_id: ModelClientId) -> None:
    """Check if the environment is set up correctly for given model client id."""
    if client_id == ModelClientId.OLLAMA and "OLLAMA" not in os.environ:
        raise ValueError(f"Required environment variable 'OLLAMA' is not set.")
    elif client_id == ModelClientId.OPENAI and "OPENAI_API_KEY" not in os.environ:
        raise ValueError(f"Required environment variable 'OPENAI_API_KEY' is not set.")
    elif client_id == ModelClientId.ANTHROPIC and "'ANTHROPIC_API_KEY'" not in os.environ:
        raise ValueError(f"Required environment variable 'ANTHROPIC_API_KEY' is not set.")



from .approver_agent import ApproverAgent
from .explorer_agent import ExplorerAgent
from .refiner_agent import RefinerAgent
from .solver_agent import SolverAgent
from .subtaskplanner_agent import SubTaskPlannerAgent
from .subtasksolver_agent import SubTaskSolverAgent
from .summarizer_agent import SummarizerAgent
from .tester_agent import TesterAgent


__all__ = [
    "ApproverAgent",
    "ExplorerAgent",
    "RefinerAgent",
    "SolverAgent",
    "SubTaskPlannerAgent",
    "SubTaskSolverAgent",
    "SummarizerAgent",
    "TesterAgent",
]



def output_truncat(output: str, output_max: int) -> (str, bool):
    """Truncating a string to a maximum length and prepends a warning message if the output exceeds that limit."""

    if len(output) > output_max:
        info_text = (f"**[ATTENTION: The output has a length of {len(output)} characters exceeded "
                     f"the maximum allowed length ({output_max} characters) and was therefore truncated. "
                     f"The section below contains the part of the output that was retained.]** \n\n")
        return f"{info_text} {output[:output_max]}", True
    return output, False



import json
import os
import shlex
from pathlib import Path
from typing import Optional

from dotenv import load_dotenv
from harbor.agents.installed.base import BaseInstalledAgent
from harbor.models.agent.context import AgentContext
from pydantic import BaseModel


class ExecInput(BaseModel):
    command: str
    cwd: str | None = None
    env: dict[str, str] | None = None
    timeout_sec: int | None = None


_AGENT_ID = "spoox-m"  # "singleton",'spoox-s','spoox-m','spoox-l'
_AGENT_ID_CHAR = "m"  # "singleton",'s','m','l'
_MODEL_CLIENT_ID = "openai"  # "ollama", "openai", "anthropic"
_MODEL_ID = "gpt-5.4-2026-03-05"  # "gpt-5.4-2026-03-05", "gpt-5.3-codex", "gpt-oss:20b","qwen3:14b","claude-sonnet-4-5","magistral:24b","gpt-5","gpt-5-mini","gpt-5-nano"

# harbor automatically copies the /logs/agent directory to local Harbor logs;
# as a result, spoox logs are automatically copied as well.
_DOCKER_LOGS_DIR = "/logs/agent/spoox"
_AGENT_MAX_TIMEOUT = 60 * 60 * 2  # 2h


class Spoox(BaseInstalledAgent):
    """
    This script defines a Harbor agent wrapper for Spoox that configures the agent with a specific model.
    Compliant with https://harborframework.com/docs/agents#installed-agents.

    run with: `harbor run -d terminal-bench@2.0 -a /benchmarks/harbor/spoox`
    run with: `harbor run -d terminal-bench@2.0 --agent-import-path "benchmarks.harbor.spoox:Spoox" -k 1`
    """

    @staticmethod
    def name() -> str:
        return f"spoox-{_AGENT_ID_CHAR}"

    @property
    def _install_agent_template_path(self) -> Path:
        """
        Path to the jinja template script for installing the spoox agent in the container.
        """
        return Path(__file__).parent / "install_spoox.sh"

    def create_run_agent_commands(self, instruction: str) -> list[ExecInput]:
        """
        Create the commands to run the agent in the container.
        For running spoox, the pre-configured venv is activated and the spoox headless command is executed.
        """
        load_dotenv()
        openai_api_key = str(os.environ.get("OPENAI_API_KEY"))
        safe_instruction = shlex.quote(instruction)
        confirmation_mode = 'no_confirmation'
        cmd = (f"spoox-h "
               f"-c {_MODEL_CLIENT_ID} -m {_MODEL_ID} -a {_AGENT_ID} -l {_DOCKER_LOGS_DIR} -x {_AGENT_MAX_TIMEOUT} "
               f"-s {confirmation_mode} -t {safe_instruction}")
        return [ExecInput(command=cmd, env={"OPENAI_API_KEY": openai_api_key})]

    def populate_context_post_run(self, context: AgentContext) -> None:
        """
        Populate the context with the results of the agent execution. Assumes the run() method has already been called.
        Parses token statistics and adds general agent configs to metadata.
        """
        meta_data = self._load_spoox_logs()
        if meta_data is None:
            return
        context.n_input_tokens = meta_data.get("model-client-total-usage-prompt-tokens")
        context.n_output_tokens = meta_data.get("model-client-total-usage-completion-tokens")
        meta_data['spoox-agent-id'] = _AGENT_ID
        meta_data['spoox-agent-id-char'] = _AGENT_ID_CHAR
        meta_data['model-client-id'] = _MODEL_CLIENT_ID
        meta_data['model-id'] = _MODEL_ID
        context.metadata = meta_data

    def _load_spoox_logs(self) -> Optional[dict]:
        """Find and load spoox logs."""
        try:
            # find spoox logs
            spoox_dir = self.logs_dir / "spoox"
            spoox_logs_dirs = [e for e in spoox_dir.iterdir() if e.name.startswith("spoox_")]
            if len(spoox_logs_dirs) != 1:
                return None
            # load metadata logs
            meta_data_logs = spoox_logs_dirs[0] / "meta_data.json"
            if not meta_data_logs.is_file():
                return None
            with meta_data_logs.open("r", encoding="utf-8") as f:
                meta_data = json.load(f)
            if isinstance(meta_data, dict):
                return meta_data
        except Exception:
            # log parsing is non-critical
            pass
        return None



from .agent_system import AgentSystem
from .errors import AgentError
from .errors import ModelClientError
from .errors import MaxOnlyTextMessagesError
from .errors import MaxIterationsError
from .mas import SpooxLarge
from .mas import SpooxMedium
from .mas import SpooxSmall
from .mas import BaseGroupChatAgent
from .singleton import SingletonAgentSystem


__all__ = [
    "AgentSystem",

    "SingletonAgentSystem",

    "BaseGroupChatAgent",
    "SpooxLarge",
    "SpooxMedium",
    "SpooxSmall",

    "AgentError",
    "ModelClientError",
    "MaxOnlyTextMessagesError",
    "MaxIterationsError",
]




import argparse
import asyncio
import copy
from pathlib import Path

import nest_asyncio
from dotenv import load_dotenv

from spoox.environment import LocalEnvironment
from spoox.environment.local_environment import ConfirmationMode
from spoox.interface import CLInterface
from spoox.utils import setup_agent_system, AgentSystemId
from spoox.environment.model_clients.utils import setup_model_client, ModelClientId
from spoox.utils_cli import CONFIG_FORM, confirm_cli_config, print_cli_header, start_loading, stop_loading, \
    print_cli_footer, print_error_message

"""
example usage:
python src/spoox/spoox_cli.py -m gpt-5-mini -a spoox-m -l False -d False -e False
"""

nest_asyncio.apply()

LOGS_DIR = Path('/tmp/spoox')
LOGS_DIR.mkdir(parents=True, exist_ok=True)


def main() -> None:
    """
    Entry point for the spoox CLI.
    This CLI script configures and runs an agent system.
    """

    parser = argparse.ArgumentParser(description="Agentu argument parser")
    parser.add_argument(
        "-c",
        "--model-client-id",
        required=False,
        help="Model client, options: 'ollama', 'openai', 'anthropic' (str).",
    )
    parser.add_argument(
        "-m",
        "--model-id",
        required=False,
        help="Model id (str)."
    )
    parser.add_argument(
        "-a",
        "--agent-id",
        required=False,
        help="Agent id (e.g. 'singleton', 'spoox-m') (str)."
    )
    parser.add_argument(
        "-l",
        "--logging",
        required=False,
        help="Show detailed logs (bool)."
    )
    parser.add_argument(
        "-s",
        "--confirmation-mode",
        required=False,
        help="The level of confirmation the agent will seek from the user before interacting with the environment, "
             "options: 'strict', 'self_evaluation', 'no_confirmation' (str).",
    )

    print_cli_header()

    # fill config
    args = parser.parse_args()
    config = copy.deepcopy(CONFIG_FORM)
    config['model_client_id']['value'] = ModelClientId(args.model_client_id) if args.model_client_id else None
    config['model_id']['value'] = str(args.model_id) if args.model_id else None
    config['agent_id']['value'] = AgentSystemId(args.agent_id) if args.agent_id else None
    config['confirmation_mode']['value'] = ConfirmationMode(args.confirmation_mode) if args.confirmation_mode else None
    if args.logging is not None:
        config['debugging_mode']['value'] = 'yes' if str(args.logging.lower()) in ("yes", "true", "t", "y") else 'no'
    config = asyncio.run(confirm_cli_config(config, LOGS_DIR))

    # extract required params from config
    client_id = ModelClientId(config['model_client_id']['value'])
    model_id = config['model_id']['value']
    agent_id = AgentSystemId(config['agent_id']['value'])
    logging_active = config['debugging_mode']['value'] == 'yes'
    confirmation_mode = ConfirmationMode(config['confirmation_mode']['value'])

    # setup agent system
    start_loading()
    load_dotenv()
    try:
        model_client = setup_model_client(client_id, model_id)
        interface = CLInterface(logging_active=logging_active)
        environment = LocalEnvironment(interface=interface, confirmation_mode=confirmation_mode)
        interface = CLInterface(logging_active)
        agent = setup_agent_system(agent_id, model_client, environment, interface, logs_dir=LOGS_DIR)
    except Exception as e:
        print_error_message(f"Exception during agent system setup:\n {str(e)}")
        return
    finally:
        stop_loading()
    print_cli_footer(config['agent_id']['value'])

    # run agent system
    try:
        asyncio.run(agent.start())
    except Exception as e:
        print_error_message(f"Exception during agent system execution:\n {str(e)}")


if __name__ == "__main__":
    main()



from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.base_agent import BaseGroupChatAgent
from spoox.agents.mas.agents.prompts import get_EXPLORER_SYSTEM_MESSAGE


class ExplorerAgent(BaseGroupChatAgent):
    """
    This agent is prompted to solely focus on exploring the environment and gathering basic information about it.
    It is not intended to solve the task, its only objective is to produce a broad overview of the given environment.
    To emphasize its purely exploratory nature, this agent is intentionally not equipped with the Python tool.
    """

    def __init__(
            self,
            topic_type: str,
            agent_system: AgentSystem,
            next_agent_topic: str,
            support_feedback: bool = False,
    ) -> None:

        system_message = get_EXPLORER_SYSTEM_MESSAGE(
            topic_type, next_agent_topic, agent_system.environment.get_additional_tool_descriptions(self), support_feedback)

        super().__init__(
            description="Agent tasked with exploring the system via its tools.",
            system_message=system_message,
            agent_system=agent_system,
            next_agent_topic_types=[next_agent_topic],
            max_internal_iterations=200,
        )



from typing import Optional

from autogen_core import CancellationToken, Component, ComponentModel
from autogen_core.code_executor import CodeExecutor
from autogen_core.tools import BaseTool
from pydantic import BaseModel, Field, model_serializer
from typing_extensions import Self

from spoox.environment.code_executors.code_executor_local import CodeBlockTimeout
from spoox.environment.tools.utils import output_truncat


"""
To handle more complex tasks effectively, equipping the agent with the ability to execute Python code can be 
particularly beneficial. While the agent could, in principle, open a Python interpreter within the Terminal tool, 
the provision of a dedicated tool that allows the definition and execution of complete Python scripts simplifies 
this process. The agent is explicitly prompted to use it for more complex sub-tasks such as data analysis or exhaustive 
algorithmic reasoning. This may also improve transparency in practice, as Python code is generally easier to read and 
likely more familiar to contemporary developers. In addition, access to a wide range of pip packages expands 
the agent’s functional capabilities.
"""


class CodeExecutionInput(BaseModel):
    """Input object for PythonTool."""

    code: str = Field(description="The Python code block that should be executed.")
    timeout: int = Field(
       description="(Optional) Maximum duration (in seconds) the code may run. "
                   "Define this only if you need to increase the default timeout of 10s. "
                   "Timeout value must not exceed 90s.",
       default=10
    )


class CodeExecutionResult(BaseModel):
    """Result object for PythonTool."""

    output: str
    exit_code: Optional[int] = None
    truncated: bool = False

    @model_serializer
    def ser_model(self) -> str:
        out = ""
        if self.exit_code is not None:
            out += f"<exit-code> {self.exit_code} </exit-code> \n"
        if self.truncated:
            out += f"<truncated-output> True </truncated-output> \n"
        out += f"<output> {self.output} </output>"
        return out


class PythonToolConfig(BaseModel):
    """Configuration for PythonTool."""

    executor: ComponentModel
    output_max: int
    description: str = "Run a Python code block once in the users current directory."


class PythonTool(BaseTool[CodeExecutionInput, CodeExecutionResult], Component[PythonToolConfig]):
    """A tool that executes Python code blocks in a code executor and returns the output."""

    component_config_schema = PythonToolConfig

    def __init__(self, executor: CodeExecutor, output_max: int = 10000):
        super().__init__(
            CodeExecutionInput,
            CodeExecutionResult,
            "PythonExecutor",
            "Run a Python code block once in the users current directory.",
            strict=True
        )
        self._executor = executor
        self._output_max = output_max

    async def run(self, args: CodeExecutionInput, cancellation_token: CancellationToken) -> CodeExecutionResult:
        # execute code
        code_block = CodeBlockTimeout(code=args.code, language="python", timeout=args.timeout)
        result = await self._executor.execute_code_blocks(
            code_blocks=[code_block], cancellation_token=cancellation_token
        )
        # make sure the output is cut when too long
        output, truncated = output_truncat(result.output, self._output_max)
        return CodeExecutionResult(exit_code=result.exit_code, output=output, truncated=truncated)

    def _to_config(self) -> PythonToolConfig:
        """Convert current instance to config object."""
        return PythonToolConfig(executor=self._executor.dump_component(), output_max=self._output_max)

    @classmethod
    def _from_config(cls, config: PythonToolConfig) -> Self:
        """Create instance from config object."""
        executor = CodeExecutor.load_component(config.executor)
        return cls(executor=executor, output_max=config.output_max)



import json
from abc import ABC, abstractmethod

from autogen_core import FunctionCall, CancellationToken, BaseAgent
from autogen_core.models import FunctionExecutionResult
from autogen_core.tools import BaseTool

from spoox.interface import Interface


class Environment(ABC):
    """
    This abstract base class manages a collection of tools and other environmental stuff,
    providing methods to start/stop/reset the environment, retrieve available tools,
    and execute tool calls with logging, error handling, and usage statistics tracking.
    Typically, an agent is equipped with an environment.
    """

    def __init__(self, interface: Interface):
        self.interface = interface
        self.tools = []
        self._started = False
        self.additional_tool_descriptions = ""

    @abstractmethod
    async def start(self):
        """Starts the environment. Should be called once during agent system startup."""
        pass

    @abstractmethod
    async def stop(self):
        """Stop the environment. Should be called once during agent system shutdown."""
        pass

    @abstractmethod
    async def reset(self):
        """
        Resets the environment. Typically called when an agent starts working in a shared environment
        to ensure no dependencies exist from previous agent operations.
        """
        pass

    @abstractmethod
    def get_tools(self, agent: BaseAgent) -> list[BaseTool]:
        """Returns a list of tools the agent should be equipped with."""
        pass

    @abstractmethod
    def get_additional_tool_descriptions(self, agent: BaseAgent) -> [str]:
        """
        Agent system prompts may include additional descriptions for their environment and tools.
        This function returns supplementary description text for the specific agent when needed.
        """
        pass

    @abstractmethod
    def _check_tool_call_confirmation(self, call: FunctionCall) -> str:
        """
        Check if user configuration should be seeked for the given FunctionCall.
        Returns an empty string if call is confirmed and can be executed, or a rejection message.
        """
        pass

    async def execute_tool_call(
            self, tools, call: FunctionCall, cancellation_token: CancellationToken, interface: Interface,
            usage_stats: dict, caller_name: str = ""
    ) -> FunctionExecutionResult:
        """
        This method executes a tool call by finding the matching tool by name, running it with the provided arguments.

        Args:
            tools (list[Tool]): List of all available tools callable by the agent.
            call (FunctionCall): FunctionCall to be executed.
            call (FunctionCall): FunctionCall to be executed.
            cancellation_token (CancellationToken): CancellationToken.
            interface (Interface): Interface for user-facing logging.
            usage_stats (dict): Dictionary of usage statistics, provided by the agent system.
            caller_name (str): Agent topic type.

        Returns:
            FunctionExecutionResult: Filled FunctionExecutionResult.
        """

        # logging
        args_parsed = json.loads(call.arguments)
        tool_name = call.name
        interface.print_tool_call(
            f"**Tool**: {tool_name}  \n**Arguments**:  \n{args_parsed}  \n",
            f"{caller_name} - tool_call"
        )

        if not self._started:
            raise RuntimeError(f"Environment must be started. Make sure `.start()` is called.")

        # find tool by name and run it
        tool = next((tool for tool in tools if tool.name == call.name), None)
        if tool is None:
            feResult = FunctionExecutionResult(call_id=call.id, content=f"Tool '{tool_name}' is not known.",
                                               is_error=True, name=call.name)

        else:
            # seek user confirmation
            rejection_message = self._check_tool_call_confirmation(call)
            if rejection_message != "":
                feResult = FunctionExecutionResult(call_id=call.id,
                                                   content=f"Tool call was rejected due to: (UserMessage) {rejection_message}.",
                                                   is_error=True, name=call.name)

            else:
                # run the tool and capture the result
                try:
                    result = await tool.run_json(args_parsed, cancellation_token)
                    feResult = FunctionExecutionResult(call_id=call.id, content=tool.return_value_as_string(result),
                                                       is_error=False, name=tool.name)
                except Exception as e:
                    feResult = FunctionExecutionResult(call_id=call.id, content=str(e), is_error=True, name=tool.name)

        # logging
        interface.print_tool_call(
            f"**Tool**: {feResult.name}  \n**Is error**: {feResult.is_error}  \n**Content**:  \n{feResult.content}  \n",
            f"{caller_name} - tool_call_result"
        )
        usage_stats['tool_calls'].append((call, feResult))
        if tool_name in usage_stats['tool_call_counts']:
            usage_stats['tool_call_counts'][tool_name] += 1
        else:
            usage_stats['tool_call_counts'][tool_name] = 1

        return feResult



from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.base_agent import BaseGroupChatAgent
from spoox.agents.mas.agents.prompts import get_SUB_TASK_SOLVER_SYSTEM_MESSAGE


class SubTaskSolverAgent(BaseGroupChatAgent):
    """
    Unlike the Solver, this agent should focus exclusively on completing the latest specified sub-task.
    It has access to the full toolset, and operates under the assumption that the sub-task has already been selected
    and sufficiently described by preceding agents.
    """

    def __init__(
            self,
            topic_type: str,
            agent_system: AgentSystem,
            planner_agent_topic_type: str,
    ) -> None:

        system_message = get_SUB_TASK_SOLVER_SYSTEM_MESSAGE(
            topic_type, planner_agent_topic_type, agent_system.environment.get_additional_tool_descriptions(self))

        super().__init__(
            description="Agent tasked to solve the given task or sub-task via its tools.",
            system_message=system_message,
            agent_system=agent_system,
            next_agent_topic_types=[planner_agent_topic_type],
            max_internal_iterations=200
        )



import html

from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel

from spoox.interface.interface import Interface, CLIColor


class LogInterfaceUserDelegate:
    """
    Delegate for storing and retrieving user interactions for preconfigured interface setup.
    Enables LogInterface to replay queued user interactions for automated testing of agent systems.
    """

    def __init__(self):
        super().__init__()
        self._user_inputs: list = []
        self._user_choices: list = []
        self.default_user_input = ""
        self.default_user_choice = ""

    @property
    def user_input(self) -> str:
        if len(self._user_inputs) == 0:
            return self.default_user_input
        return self._user_inputs.pop(0)

    @property
    def user_choice(self) -> str:
        if len(self._user_choices) == 0:
            return self.default_user_choice
        return self._user_choices.pop(0)

    @user_input.setter
    def user_input(self, inp: str or list[str]) -> None:
        if isinstance(inp, str):
            self._user_inputs.append(inp)
        if isinstance(inp, list):
            self._user_inputs += inp

    @user_choice.setter
    def user_choice(self, choice: str or list[str]) -> None:
        if isinstance(choice, str):
            self._user_choices.append(choice)
        if isinstance(choice, list):
            self._user_choices += choice


class LogInterface(Interface):
    """
    Interface for agent systems that uses a LogInterfaceUserDelegate for preconfigured user interactions.
    Furthermore, all operations are automatically logged.
    """

    console = Console()

    def __init__(self, logging_active: bool = False, print_live: bool = False):
        super().__init__(logging_active)
        self.logs = []
        self.user_delegate = LogInterfaceUserDelegate()
        self.print_live = print_live

    def print(self, out_text: str, title: str = "", color: CLIColor = CLIColor.DEFAULT) -> None:
        """Prints out_text to the interface with optional title and color parameters."""
        if self.print_live:
            md = Markdown(html.escape(out_text))  # html.escape shows markdown with html tags
            panel = Panel(md, title=title, style=color.value)
            self.console.print(panel)
        self.logs.append((f"{title}", out_text, color.value))

    def request_user_input(self, query: str, default: str = "", allow_empty_input: bool = False) -> str:
        """
        Requests arbitrary text input from the user.
        The next queued user input is taken from the user_delegate.
        """
        self.logs.append(("user_input_request", query))
        user_input = self.user_delegate.user_input
        self.print_user_message(user_input)
        return user_input

    def request_select_choice(self, question: str, choices: list[str]) -> str:
        """
        Requests the user to select one item from choices.
        The next queued user choice is taken from the user_delegate.
        """
        self.logs.append(("select_choice_request_question", question))
        self.logs.append(("select_choice_request_choices", ', '.join(choices)))
        user_choice = self.user_delegate.user_choice
        self.logs.append(("selected_choice", user_choice))
        return user_choice

    def reset(self) -> None:
        """Resets the interface."""
        self.logs = []
        self.user_delegate = LogInterfaceUserDelegate()



import asyncio
import random

from autogen_core import CancellationToken, Component
from autogen_core.tools import BaseTool
from pydantic import BaseModel, Field, model_serializer
from typing_extensions import Self

from spoox.environment.code_executors.tmux_terminal_session import TmuxTerminalSession


"""
Executing isolated Bash commands does not grant the agent the same level of control over the terminal as a human has. 
Many terminal applications, such as vim, top, or man offer their own interactive CLIs, which cannot be meaningfully 
accessed through single Bash calls. To enable successful interaction with these interfaces, the agent needs a tool
that maintains a persistent shell instance and supports sending actual keyboard inputs such as Ctrl-C or Enter.
To address this need, we have implemented a dedicated Terminal tool. Although we explored various designs, 
ranging from multi-parameter invocation schemes to more guided action spaces, the approach that worked best in
practice was, once again, the simplest one. The Terminal tool maintains an persistent terminal session using the 
widely adopted tmux package. Tmux enables the creation of detachable, persistent terminal sessions with panes and 
windows for structured command-line workflows, all controllable via tmux commands. The agent calls the tool by defining 
a command as a string and, optionally, a boolean parameter named enter. By default, enter is set to true, meaning
that after the specified command is typed in, an Enter keystroke is issued to submit it. The agent must explicitly set 
enter to false when it wants to input text without submitting it immediately. The latter is typically required when an 
interactive terminal interface only needs a single character such as q to exit, or when the agent intends to send 
specific keystrokes like Ctrl-b, Tab, or Up. On each invocation, the tool returns the name of the program currently 
running, typically bash in the default terminal. In addition, not only the last output is returned, instead, the tool 
provides a full medium sized terminal window with 128 columns and 16 rows. This allows the agent to view the complete 
terminal screen and interact with CLIs.
"""


class TerminalInput(BaseModel):
    """Input object for TerminalInput."""

    command: str = Field(description="The command that should be executed in the persistent terminal instance.")
    enter: bool = Field(
        default=True,
        description="Set to True to simulate pressing the Enter key after the command, ensuring it is submitted to the terminal. "
                    "Set to False if the command should only be typed into the terminal without executing Enter. "
                    "Typically, set this to True when submitting Bash commands, and to False when sending tmux keys (e.g., C-b, Tab, Up). "
                    "This field is optional, the default is set to True.")


class TerminalResult(BaseModel):
    """Result object for TerminalTool."""

    running_program: str
    current_screen: str

    @model_serializer
    def ser_model(self) -> str:
        return (f"<running-program> {self.running_program} </running-program>  "
                f"\n  <current-screen>  \n  {self.current_screen}  \n  </current-screen>")


class TerminalToolConfig(BaseModel):
    """Configuration for TerminalTool"""

    tmux_session_name: str
    description: str = (
        "Executes the command in a persistent terminal session, waits for 2 seconds and returns the updated visible terminal screen buffer (size: 128 columns and 16 rows). "
        "Supports both shell commands and interactive commands that follow the tmux tokens format; "
        "tmux token examples: C-a, C-b, C-c, C-l, Tab, Space, Home, End, Insert, Delete, Up, Down, Left, Right, Escape. "
        "Especially ideal for running and controlling interactive terminal programs such as 'git log' or 'vim'."
    )


class TerminalTool(BaseTool[TerminalInput, TerminalResult], Component[TerminalToolConfig]):
    """A tool that runs an active terminal can execute commands and returns the entire terminal screen."""

    component_config_schema = TerminalToolConfig

    def __init__(self, tmux_session_name: str = None):
        super().__init__(
            TerminalInput,
            TerminalResult,
            "Terminal",
            (
                "Executes the command in a persistent terminal session, waits for 2 seconds and returns the updated visible terminal screen buffer (size: 128 columns and 16 rows). "
                "Supports both shell commands and interactive commands that follow the tmux tokens format; "
                "tmux token examples: C-a, C-b, C-c, C-l, Tab, Space, Home, End, Insert, Delete, Up, Down, Left, Right, Escape. "
                "Especially ideal for running and controlling interactive terminal programs such as 'git log' or 'vim'."
            ),
            strict=True
        )
        self.tmux_session_name = None
        self._session = None
        self._init_session(tmux_session_name)

    def _init_session(self, tmux_session_name: str = None) -> None:
        """Init a new tmux session - also overrides the current tmux_session_name."""

        if tmux_session_name is None:
            rand_suffix = str(random.randint(10000000, 99999999))
            self.tmux_session_name = f"ts-{rand_suffix}"
        else:
            self.tmux_session_name = tmux_session_name
        self._session = TmuxTerminalSession(session_name=self.tmux_session_name)
        self._session.clear_history()

    async def run(self, args: TerminalInput, cancellation_token: CancellationToken = None) -> TerminalResult:
        # add enter key stroke
        if args.enter:
            self._session.send_line(args.command)
        else:
            self._session.send_keys(args.command)
        # before getting the screen -> wait for 2s
        await asyncio.sleep(2)
        current_screen, running_program = self._session.get_screen()
        return TerminalResult(
            running_program=running_program,
            current_screen=current_screen
        )

    async def reset(self):
        if self._session is not None:
            self._session.kill_session()
        self._init_session()

    async def stop(self):
        if self._session is not None:
            self._session.kill_session()

    def _to_config(self) -> TerminalToolConfig:
        """Convert current instance to config object."""
        return TerminalToolConfig(tmux_session_name=self.tmux_session_name)

    @classmethod
    def _from_config(cls, config: TerminalToolConfig) -> Self:
        """Create instance from config object."""
        return cls(tmux_session_name=config.tmux_session_name)



import subprocess
import time


class TmuxTerminalSession:
    """
    Control a persistent tmux pane from Python.

    - A tmux session is created if it doesn't exist.
    - One can send commands / key-presses.
    - One can read the current visible screen contents.
    """

    def __init__(self, session_name="terminal", window_index=0, pane_index=0):
        self.session_name = session_name
        self.window_index = window_index
        self.pane_index = pane_index
        self.target = f"{session_name}:{window_index}.{pane_index}"
        self._ensure_session()

    def _tmux(self, *args, capture_output=True, text=True, check=True):
        """Run a tmux command and return the CompletedProcess."""
        return subprocess.run(
            ["tmux", *args],
            capture_output=capture_output,
            text=text,
            check=check,
        )

    def _ensure_session(self):
        """Create the session if it does not exist yet."""
        result = subprocess.run(
            ["tmux", "has-session", "-t", self.session_name],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        if result.returncode != 0:
            # Create a detached session running your shell (default: $SHELL)
            self._tmux("new-session", "-d", "-s", self.session_name)
            time.sleep(2)  # todo can be done better
            # set a fixed terminal size
            self._tmux("resize-window", "-t", self.session_name, "-x", "128", "-y", "16")

    def send_line(self, line: str):
        """Send a command followed by Enter (like typing a command and pressing Return)."""
        self._tmux("send-keys", "-t", self.target, line, "Enter")

    def send_keys(self, *keys: str):
        """Send arbitrary keys (no implicit Enter). You can pass tmux key tokens (e.g. 'C-c') or literal strings."""
        self._tmux("send-keys", "-t", self.target, *keys)

    def get_screen(self) -> (str, str):
        """
        Get the *current visible* screen contents of the pane as text and the currently *running program*.

        If you want escape sequences (colors, cursor pos, etc),
        call capture-pane with -e instead and pipe through a terminal emulator.
        """
        # Get the current command
        cmd_proc = self._tmux(
            "display-message",
            "-p",
            "-F",
            "#{pane_current_command}",
            "-t",
            self.target,
        )
        current_cmd = cmd_proc.stdout.strip()

        # Ask tmux for the pane height
        height_proc = self._tmux(
            "display-message",
            "-p",
            "-F",
            "#{pane_height}",
            "-t",
            self.target,
        )
        height = int(height_proc.stdout.strip())

        # Capture the last <height> lines (i.e. full visible screen)
        cap = self._tmux(
            "capture-pane",
            "-t",
            self.target,
            "-p",  # print to stdout
            "-S", f"-{height}",
        )
        screen_text = cap.stdout.rstrip("\n")

        return screen_text, current_cmd

    def clear_history(self):
        """Clear terminal screen history."""
        self.send_line("clear")
        time.sleep(1)  # todo can be done better
        self._tmux("clear-history", "-t", self.target)

    def kill_session(self):
        """Kill the whole tmux session (optional)."""
        self._tmux("kill-session", "-t", self.session_name)



from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.base_agent import BaseGroupChatAgent
from spoox.agents.mas.agents.prompts import get_TESTER_SYSTEM_MESSAGE


class TesterAgent(BaseGroupChatAgent):
    """
    The sole purpose of the Tester Agent is to actively test the implemented solution.
    Thereby, it should not fix any issues, instead, it writes tests, executes them,
    and provides a natural language summary of its findings, using all coding tools.
    """

    def __init__(
            self,
            topic_type: str,
            agent_system: AgentSystem,
            previous_agent_topic_type: str,
            next_agent_topic_type: str,
    ) -> None:

        system_message = get_TESTER_SYSTEM_MESSAGE(
            topic_type,
            previous_agent_topic_type,
            next_agent_topic_type,
            agent_system.environment.get_additional_tool_descriptions(self)
        )

        super().__init__(
            description="Agent tasked to test the solution of the given task.",
            system_message=system_message,
            agent_system=agent_system,
            next_agent_topic_types=[previous_agent_topic_type, next_agent_topic_type],
            max_internal_iterations=200,
        )



from typing import Sequence, Optional

from autogen_core import CancellationToken, FunctionCall
from autogen_core.models import LLMMessage, RequestUsage
from autogen_core.tools import Tool, ToolSchema
from openai import OpenAI
from openai.types.responses import FunctionToolParam

from spoox.environment.model_clients.custom_response_types import CreateResultOpenAI


class CustomOpenAIResponseAPIClient:
    # todo cleanup

    def __init__(self, model_id: str):

        self._client = OpenAI()
        self._model_id = model_id
        self.model_info = {'model_id': model_id}
        self.prompt_tokens = 0
        self.completion_tokens = 0

    def actual_usage(self) -> RequestUsage:
        return RequestUsage(
            prompt_tokens=self.prompt_tokens,
            completion_tokens=self.completion_tokens
        )

    def total_usage(self) -> RequestUsage:
        return self.actual_usage()

    async def create(
            self,
            messages: Sequence[LLMMessage],
            tools: Sequence[Tool | ToolSchema] = [],
            cancellation_token: Optional[CancellationToken] = None,
    ) -> CreateResultOpenAI:

        # parse messages
        parsed_messages = list()
        for m in messages:

            if m.type == 'UserMessage':
                parsed_messages.append({
                    "type": "message",
                    "role": "user",
                    "content": m.content,
                })
            elif m.type == 'SystemMessage':
                parsed_messages.append({
                    "type": "message",
                    "role": "developer",
                    "content": m.content,
                })
            elif m.type == 'AssistantMessageOpenAI':
                parsed_messages.extend(m.response_items)
            elif m.type == 'AssistantMessage' and isinstance(m.content, str):
                # if it is an AssistantMessage and no AssistantMessageOpenAI -> it always only contains
                # the summary of the previous agent
                parsed_messages.append({
                    "type": "message",
                    "role": "assistant",
                    "content": m.content,
                    "phase": "commentary"
                })
            elif m.type == 'FunctionExecutionResultMessage':
                for fR in m.content:
                    parsed_messages.append({
                        "type": "function_call_output",
                        "call_id": fR.call_id,
                        "output": fR.content,
                    })
            else:
                raise ValueError(f"Unexpected message type: {m.type}")

        # parse tools
        parsed_tools = list()
        for t in tools:
            t_schema = t.schema
            tool: FunctionToolParam = {
                "type": "function",
                "name": t_schema["name"],
                "description": t_schema["description"],
                "parameters": t_schema["parameters"],
                "strict": t_schema.get("strict", True),
            }
            parsed_tools.append(tool)

        # request model
        response = self._client.responses.create(
            model=self._model_id,
            input=parsed_messages,
            tools=parsed_tools,
            reasoning={"effort": "high"},
            store=True,
            max_output_tokens=40000,
            parallel_tool_calls=False,
        )

        # parse token usage
        prompt_tokens = 0
        completion_tokens = 0
        if response.usage is not None:
            prompt_tokens += response.usage.input_tokens
            completion_tokens += response.usage.output_tokens
        self.prompt_tokens += prompt_tokens
        self.completion_tokens += completion_tokens

        if response.status == "incomplete" and response.incomplete_details.reason == "max_output_tokens":
            print("Ran out of tokens")  # todo check and delete


        # parse response
        func_calls: list[FunctionCall] = list()
        output_texts: list[str] = list()
        thoughts: list[str] = list()
        for item in response.output:
            if item.type == "message":
                output_texts.extend(i.text for i in item.content or [] if i.type == "output_text")
            elif item.type == "reasoning":
                thoughts.extend(i.text for i in item.content or [])
            elif item.type == "function_call":
                func_calls.append(FunctionCall(
                    id=item.call_id,
                    arguments=item.arguments,
                    name=item.name,
                ))
            else:
                raise ValueError(f"Unknown response.output type {item.type} - {item}")

        if len(func_calls) > 0:
            return CreateResultOpenAI(
                finish_reason="function_calls",
                content=func_calls,
                usage=RequestUsage(
                    prompt_tokens=prompt_tokens,
                    completion_tokens=completion_tokens,
                ),
                cached=False,
                thought=';\n'.join(output_texts + thoughts),
                response_items=response.output
            )
        return CreateResultOpenAI(
            finish_reason="stop",
            content=';\n'.join(output_texts),
            usage=RequestUsage(
                prompt_tokens=prompt_tokens,
                completion_tokens=completion_tokens,
            ),
            cached=False,
            thought=';\n'.join(thoughts),
            response_items=response.output
        )



from anthropic import BaseModel
from autogen_core.models import UserMessage, AssistantMessage

# generic topic type used for group chat messages.
GROUP_CHAT_TOPIC_TYPE = "groupchat"


# all messages include 'nonce' to ensure each message is unique,
# preventing handlers from merging messages sent close together in time.

class GroupChatMessage(BaseModel):
    """Text message for a group chat. Typically distributed to all agents."""
    nonce: str
    body: UserMessage | AssistantMessage


class RequestToSpeak(BaseModel):
    """Organizational message requesting an agent to start working."""
    nonce: str



import html

import questionary
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel

from spoox.interface.interface import Interface, CLIColor


class CLInterface(Interface):
    """
    Simple CLI interface for agent systems.
    Prints outputs and requests inputs one below the other in the command line.
    """

    console = Console()

    def __init__(self, logging_active: bool = False):
        super().__init__(logging_active)

    def print(self, out_text: str, title: str = "", color: CLIColor = CLIColor.DEFAULT) -> None:
        """Prints out_text to the interface with optional title and color parameters."""
        md = Markdown(html.escape(out_text))  # html.escape shows markdown with html tags
        panel = Panel(md, title=title, style=color.value)
        self.console.print(panel)

    def request_user_input(self, query: str, default: str = "", allow_empty_input: bool = False) -> str:
        """Requests arbitrary text input from the user."""
        self.console.print("")
        if allow_empty_input:
            user_input = questionary.text(query, qmark='👻 ', default=default).ask()
        else:
            user_input = questionary.text(query, qmark='👻 ', default=default, validate=lambda s: bool(s)).ask()
        self.console.print("")
        return user_input

    def request_select_choice(self, question: str, choices: list[str]) -> str:
        """Requests the user to select one item from choices."""
        self.console.print("")
        user_choice = questionary.select(question, choices, qmark='👻 ').ask()
        self.console.print("")
        return user_choice

    def reset(self) -> None:
        """Resets the interface."""
        pass



from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.base_agent import BaseGroupChatAgent
from spoox.agents.mas.agents.prompts import get_SOLVER_SYSTEM_MESSAGE


class SolverAgent(BaseGroupChatAgent):
    """
    As its name implies, this agent is responsible for actively solving the task.
    It continues working until the task is completed or no further meaningful contributions are possible.
    The agent is equipped with all available tools to ensure the maximum possible action space.
    """

    def __init__(
            self,
            topic_type: str,
            agent_system: AgentSystem,
            tester_agent_topic_type: str,
    ) -> None:

        system_message = get_SOLVER_SYSTEM_MESSAGE(
            topic_type, tester_agent_topic_type, agent_system.environment.get_additional_tool_descriptions(self))

        super().__init__(
            description="Agent tasked to solve the given task via its tools.",
            system_message=system_message,
            agent_system=agent_system,
            next_agent_topic_types=[tester_agent_topic_type],
            max_internal_iterations=100,
        )



from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.base_agent import BaseGroupChatAgent
from spoox.agents.mas.agents.prompts import get_REFINER_SYSTEM_MESSAGE


class RefinerAgent(BaseGroupChatAgent):
    """
    Equipped with all tools, this agent is responsible for resolving discovered bugs
    and refining or completing the existing implementation of the overall task.
    It assumes that previous agents have already identified bugs, missing components, and other errors in the solution.
    Its sole focus is to address and fix these most recent findings.
    """

    def __init__(
            self,
            topic_type: str,
            agent_system: AgentSystem,
            tester_topic_type: str,
            approver_topic_type: str,
    ) -> None:

        system_message = get_REFINER_SYSTEM_MESSAGE(
            topic_type, tester_topic_type, approver_topic_type, agent_system.environment.get_additional_tool_descriptions(self))

        super().__init__(
            description="Agent tasked to refine and fix the implemented task solution.",
            system_message=system_message,
            agent_system=agent_system,
            next_agent_topic_types=[tester_topic_type, approver_topic_type],
            max_internal_iterations=200
        )




class AgentError(Exception):
    """Base exception for all agent errors."""

    def __init__(self, agent_id: str, message: str):
        m = f"Agent exception - for agent: {agent_id} - {message}"
        super().__init__(m)
        self.agent_id = agent_id


class ModelClientError(AgentError):
    """Model client error after maximum retry limit is reached."""

    def __init__(self, agent_id: str, max_retrials: int, model_client_exception: Exception):
        super().__init__(
            agent_id,
            f"Model client exception:\n{str(model_client_exception)} and max reached ({max_retrials})"
        )
        self.max_retrials = max_retrials
        self.model_client_exception = model_client_exception


class MaxOnlyTextMessagesError(AgentError):
    """The maximum limit of consecutive text-only messages has been reached."""

    def __init__(self, agent_id: str, max_only_text_messages: int):
        super().__init__(
            agent_id, f"Max only text messages reached ({max_only_text_messages})")
        self.max_only_text_messages = max_only_text_messages


class MaxIterationsError(AgentError):
    """The maximum number of agent internal iterations has been reached."""

    def __init__(self, agent_id: str, max_iterations: int):
        super().__init__(agent_id, f"Max agent iterations reached ({max_iterations})")
        self.max_iterations = max_iterations



from abc import ABC, abstractmethod
from enum import Enum


class CLIColor(Enum):
    """Collection of standard colors used in the interfaces."""

    DEFAULT = '#909090'
    ORANGE = '#ab3e03'
    DARKORANGE = '#70441a'
    GREY = '#555555'
    DARKGREY = '#3d3d3d'
    CYAN = '#008B8B'
    BLUE = '#193754'
    GREEN = '#2C5C27'
    LILA = '#2D1F5E'
    RED = '#521A20'


class Interface(ABC):
    """
    Abstract base class that defines a standardized I/O interface.
    Represents the main interface for the end user.
    Concrete implementations of this interface are compatible with any agent systems.
    """

    def __init__(self, logging_active: bool = False):
        self.logging_active = logging_active

    @abstractmethod
    def print(self, out_text: str, title: str = "", color: CLIColor = CLIColor.DEFAULT) -> None:
        """Prints out_text to the interface with optional title and color parameters."""
        pass

    def print_highlight(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.ORANGE)

    def print_shadow(self, out_text: str, title: str = "") -> None:
        self.print(out_text, title, CLIColor.GREY)

    def print_thought(self, out_text: str, title: str = "thought") -> None:
        self.print(out_text, title, CLIColor.CYAN)

    def print_tool_call(self, out_text: str, title: str = "tool call") -> None:
        self.print(out_text, title, CLIColor.DARKORANGE)

    def print_user_message(self, out_text: str, title: str = "user input") -> None:
        self.print(out_text, title, CLIColor.LILA)

    def print_logging(self, out_text: str, title: str = "logs") -> None:
        if self.logging_active:
            self.print(out_text, title, CLIColor.DARKGREY)

    @abstractmethod
    def request_user_input(self, query: str, default: str = "", allow_empty_input: bool = False) -> str:
        """Requests arbitrary text input from the user."""
        pass

    @abstractmethod
    def request_select_choice(self, question: str, choices: list[str]) -> str:
        """Requests the user to select one item from choices."""
        pass

    @abstractmethod
    def reset(self) -> None:
        """Resets the interface."""
        pass



import asyncio
import re
from typing import List, Optional, Tuple

from autogen_core import RoutedAgent, message_handler, MessageContext, FunctionCall
from autogen_core.models import SystemMessage, LLMMessage, AssistantMessage, \
    FunctionExecutionResultMessage, CreateResult

from spoox.agents.agent_system import AgentSystem
from spoox.agents.errors import ModelClientError, MaxOnlyTextMessagesError, MaxIterationsError, AgentError
from spoox.agents.singleton.messages import PublicMessage
from spoox.agents.singleton.prompts import get_SINGLETON_SYSTEM_PROMPT


# to ensure progress and prevent endless text-only replies, a limit on consecutive text-only messages is enforced
MAX_ONLY_TEXT_MESSAGES = 3

# if the model client encounters errors, the agent allows a maximum of three retry attempts
MAX_MODEL_CLIENT_ERRORS_RETRIALS = 3


class SingletonAgent(RoutedAgent):
    """
    This agent represents the single-agent setup and is prompted to solve the entire task on its own. It is equipped
    with all available tools, and to provide additional guidance, we included a brief sequential abstraction
    of a typical problem-solving procedure consisting of the steps: explore, plan, solve, test, and summarize.

    As soon as it receives a RequestToSpeak, it starts working by calling tools and reasoning aloud
    until it finishes by including the finished_tag in its response.
    """

    finished_tag = "finished"

    def __init__(self, agent_system: AgentSystem, max_internal_iterations: int = 100) -> None:

        super().__init__(description="Single agent responsible for handling and completing the entire task.")
        self._max_internal_iterations = max_internal_iterations

        self._environment = agent_system.environment
        self._model_client = agent_system.model_client
        self._interface = agent_system.interface
        self._usage_stats = agent_system.usage_stats
        self._save_logs_f = agent_system.save_logs
        self._return_next_time_possible_event = agent_system.timeout_event
        self._tools = self._environment.get_tools(self)

        system_message = get_SINGLETON_SYSTEM_PROMPT(
            self.finished_tag, self._environment.get_additional_tool_descriptions(self))
        self._chat_history: List[LLMMessage] = [SystemMessage(content=system_message)]

        # logging
        self._interface.print_logging(system_message, f"logging - {self.id.type} - system_message")
        for t in self._tools:
            self._interface.print_logging(str(t.schema), f"logging - {self.id.type} - tool_schema")

    @message_handler
    async def handle_request_to_speak(self, message: PublicMessage, ctx: MessageContext) -> None:
        """Agent is requested to speak: internal execution loop is started."""

        try:
            self._chat_history.append(message.body)
            await self._agent_loop(ctx)
        except AgentError as e:
            self._interface.print_highlight(str(e), "Agent Error")
            self._usage_stats["agent_errors"].append(e)
        except Exception as e:
            self._interface.print_highlight(str(e), "Unexpected Error")
            self._usage_stats["agent_errors"].append(e)

    async def _agent_loop(self, ctx: MessageContext):
        """Request the LLM, process its response, and repeat until the agent has finished."""

        # tracking consecutive model client errors and LLM "only-text" responses
        counter_only_text_messages = 0
        model_client_errors = 0

        for i in range(1, self._max_internal_iterations + 1):

            # handling agent system timeout event
            if self._return_next_time_possible_event.is_set():
                return

            # logging
            self._save_logs_f()
            self._usage_stats['llm_calls_count'] += 1

            # request model client (llm)
            llm_res, model_client_errors = await self._request_llm(ctx, model_client_errors)
            if llm_res is None:
                continue
            self._interface.print_logging(str(llm_res), f"logging - {self.id.type} - entire llm_res")
            content = llm_res.content

            # add the response to session and print thoughts if available
            self._chat_history.append(
                AssistantMessage(content=content, thought=llm_res.thought, source=self.id.type))
            if llm_res.thought:
                self._interface.print_thought(llm_res.thought)

            # check if just text response
            if isinstance(content, str):
                self._interface.print(content, f"{self.id.type} - message")
                # check if `finished_tag` is included
                patter = rf"\[[^\]]*{re.escape(self.finished_tag)}[^\]]*\]"
                if re.search(patter, content, flags=re.IGNORECASE):
                    return
                # check if MAX_ONLY_TEXT_MESSAGES is reached
                counter_only_text_messages += 1
                if counter_only_text_messages > MAX_ONLY_TEXT_MESSAGES:
                    raise MaxOnlyTextMessagesError(self.id.type, MAX_ONLY_TEXT_MESSAGES)
                continue

            # check if tool calls (if it is not string it has to be a list of tool calls)
            assert isinstance(content, list) and all(isinstance(call, FunctionCall) for call in content)
            counter_only_text_messages = 0
            tool_results_message = await self._exec_tools(ctx, content)
            self._chat_history.append(tool_results_message)

        raise MaxIterationsError(self.id.type, self._max_internal_iterations)

    async def _request_llm(self, ctx: MessageContext, model_client_errors: int) -> Tuple[Optional[CreateResult], int]:
        """Invokes the model client (LLM) and handles any exception that occurs."""

        try:
            llm_res = await self._model_client.create(
                messages=self._chat_history,
                tools=self._tools,
                cancellation_token=ctx.cancellation_token,
            )
        except Exception as e:
            self._usage_stats['model_client_exceptions'].append(e)
            self._interface.print_logging(str(e), "Model Client Error")
            if model_client_errors >= MAX_MODEL_CLIENT_ERRORS_RETRIALS:
                raise ModelClientError(self.id.type, MAX_MODEL_CLIENT_ERRORS_RETRIALS, e)
            else:
                return None, model_client_errors + 1
        # llm call success
        self._usage_stats['prompt_tokens'].append(llm_res.usage.prompt_tokens)
        self._usage_stats['completion_tokens'].append(llm_res.usage.completion_tokens)
        return llm_res, 0

    async def _exec_tools(self, ctx: MessageContext, calls: list[FunctionCall]) -> FunctionExecutionResultMessage:
        """Executes available tool calls."""

        tool_results = await asyncio.gather(
            *[
                self._environment.execute_tool_call(
                    self._tools, c, ctx.cancellation_token, self._interface, self._usage_stats, self.id.type
                )
                for c in calls
            ]
        )
        return FunctionExecutionResultMessage(content=tool_results)



from enum import Enum
from pathlib import Path

from autogen_core.models import ChatCompletionClient

from spoox.agents import AgentSystem
from spoox.agents import SingletonAgentSystem
from spoox.agents import SpooxLarge
from spoox.agents import SpooxMedium
from spoox.agents import SpooxSmall
from spoox.environment import Environment
from spoox.environment.model_clients.custom_client_openai import CustomOpenAIResponseAPIClient
from spoox.interface import Interface


class AgentSystemId(Enum):
    """All available agent system ids."""

    SINGLETON = 'singleton'
    SPOOX_S = 'spoox-s'
    SPOOX_M = 'spoox-m'
    SPOOX_L = 'spoox-l'


def setup_agent_system(agent_system_id: AgentSystemId,
                       model_client: ChatCompletionClient | CustomOpenAIResponseAPIClient,
                       environment: Environment,
                       interface: Interface,
                       timeout: int = 600,
                       logs_dir: Path = Path.cwd()) -> AgentSystem:
    """Based on the provided 'agent_id', create the corresponding agent system instance."""

    if agent_system_id == AgentSystemId.SINGLETON:
        return SingletonAgentSystem(interface, model_client, environment, timeout, logs_dir)
    if agent_system_id == AgentSystemId.SPOOX_S:
        return SpooxSmall(interface, model_client, environment, timeout, logs_dir)
    if agent_system_id == AgentSystemId.SPOOX_M:
        return SpooxMedium(interface, model_client, environment, timeout, logs_dir)
    if agent_system_id == AgentSystemId.SPOOX_L:
        return SpooxLarge(interface, model_client, environment, timeout, logs_dir)
    raise ValueError(f"Selected agent system '{agent_system_id}' not known.")



from typing import List, Literal

from autogen_core.models import AssistantMessage, CreateResult
from openai.types.responses import ResponseOutputItem


class AssistantMessageOpenAI(AssistantMessage):
    """AutoGen AssistantMessage extended by openai response API context."""
    response_items: List[ResponseOutputItem]
    type: Literal["AssistantMessageOpenAI"] = "AssistantMessageOpenAI"


class CreateResultOpenAI(CreateResult):
    """AutoGen CreateResult extended by openai response API context."""
    response_items: List[ResponseOutputItem]



import asyncio
import uuid
from pathlib import Path

from autogen_core import DefaultTopicId
from autogen_core import TypeSubscription
from autogen_core.models import UserMessage, ChatCompletionClient

from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.messages import GroupChatMessage, RequestToSpeak, GROUP_CHAT_TOPIC_TYPE
from spoox.agents.mas.agents import SolverAgent
from spoox.agents.mas.agents import SummarizerAgent
from spoox.agents.mas.agents import TesterAgent
from spoox.environment import Environment
from spoox.environment.model_clients.custom_client_openai import CustomOpenAIResponseAPIClient
from spoox.interface import Interface


class SpooxSmall(AgentSystem):
    """
    This agent system implementation is based on the spoox-s multi-agent architecture,
    described in the spoox scaling study.
    It consists of three agents and a single feedback loop.
    """

    # all topic types
    solver_topic_type = "solver"
    tester_topic_type = "tester"
    summarizer_topic_type = "summarizer"

    def __init__(self, interface: Interface, model_client: ChatCompletionClient | CustomOpenAIResponseAPIClient,
                 environment: Environment, timeout: int = 600, logs_dir: Path = Path.cwd()):
        super().__init__(interface, model_client, environment, timeout, logs_dir)
        # agents
        self._solver_agent = None
        self._tester_agent = None
        self._summarizer_agent = None

    async def _build_agents(self):
        """Initializing all agents, including all message subscriptions."""

        self._solver_agent = await SolverAgent.register(
            self.runtime,
            self.solver_topic_type,
            lambda: SolverAgent(
                topic_type=self.solver_topic_type,
                agent_system=self,
                tester_agent_topic_type=self.tester_topic_type,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.solver_topic_type, agent_type=self._solver_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._solver_agent.type))

        self._tester_agent = await TesterAgent.register(
            self.runtime,
            self.tester_topic_type,
            lambda: TesterAgent(
                topic_type=self.tester_topic_type,
                agent_system=self,
                previous_agent_topic_type=self.solver_topic_type,
                next_agent_topic_type=self.summarizer_topic_type,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.tester_topic_type, agent_type=self._tester_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._tester_agent.type))

        self._summarizer_agent = await SummarizerAgent.register(
            self.runtime,
            self.summarizer_topic_type,
            lambda: SummarizerAgent(
                topic_type=self.summarizer_topic_type,
                agent_system=self,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.summarizer_topic_type, agent_type=self._summarizer_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._summarizer_agent.type))

    async def _trigger_agents(self, user_input: str) -> None:
        """Triggers the execution flow of the agent system's single agents, given the latest user input."""

        await self.runtime.publish_message(
            message=GroupChatMessage(nonce=str(uuid.uuid4()), body=UserMessage(content=user_input, source="user")),
            topic_id=DefaultTopicId(type=GROUP_CHAT_TOPIC_TYPE)
        )
        # 0.1 delay to ensure the GroupChatMessage can be observed before the RequestToSpeak
        # (I think it is not required, however, it certainly does not hurt)
        await asyncio.sleep(0.1)
        await self.runtime.publish_message(
            message=RequestToSpeak(nonce=str(uuid.uuid4())),
            topic_id=DefaultTopicId(type=self.solver_topic_type)
        )

    def get_state(self):
        """Returns the current state of the agent system for logging and later analysis."""
        return {
            'solver_agent': self._solver_agent,
            'tester_agent': self._tester_agent,
            'summarizer_agent': self._summarizer_agent,
        }



import asyncio
import uuid
from pathlib import Path

from autogen_core import DefaultTopicId
from autogen_core import TypeSubscription
from autogen_core.models import UserMessage, ChatCompletionClient

from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.messages import GroupChatMessage, RequestToSpeak, GROUP_CHAT_TOPIC_TYPE
from spoox.agents.mas.agents import ApproverAgent
from spoox.agents.mas.agents import ExplorerAgent
from spoox.agents.mas.agents import SubTaskSolverAgent
from spoox.agents.mas.agents import SubTaskPlannerAgent
from spoox.agents.mas.agents import SummarizerAgent
from spoox.agents.mas.agents import TesterAgent
from spoox.agents.mas.agents import RefinerAgent
from spoox.environment import Environment
from spoox.environment.model_clients.custom_client_openai import CustomOpenAIResponseAPIClient
from spoox.interface import Interface


class SpooxLarge(AgentSystem):
    """
    This agent system implementation is based on the spoox-l multi-agent architecture,
    described in the spoox scaling study.
    It consists of seven agents and five feedback loops.
    """

    # all topic types
    explorer_topic_type = "explorer"
    sub_task_planner_topic_type = "subtaskplanner"
    sub_task_solver_topic_type = "subtasksolver"
    tester_topic_type = "tester"
    refiner_topic_type = "refiner"
    approver_topic_type = "approver"
    summarizer_topic_type = "summarizer"

    def __init__(self, interface: Interface, model_client: ChatCompletionClient | CustomOpenAIResponseAPIClient,
                 environment: Environment, timeout: int = 600, logs_dir: Path = Path.cwd()):
        super().__init__(interface, model_client, environment, timeout, logs_dir)
        # agents
        self._explorer_agent = None
        self._sub_task_planner_agent = None
        self._sub_task_solver_agent = None
        self._tester_agent = None
        self._refiner_agent = None
        self._approver_agent = None
        self._summarizer_agent = None

    async def _build_agents(self):
        """Initializing all agents, including all message subscriptions."""

        self._explorer_agent = await ExplorerAgent.register(
            self.runtime,
            self.explorer_topic_type,
            lambda: ExplorerAgent(
                topic_type=self.explorer_topic_type,
                agent_system=self,
                next_agent_topic=self.sub_task_planner_topic_type,
                support_feedback=True,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.explorer_topic_type, agent_type=self._explorer_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._explorer_agent.type))

        self._sub_task_planner_agent = await SubTaskPlannerAgent.register(
            self.runtime,
            self.sub_task_planner_topic_type,
            lambda: SubTaskPlannerAgent(
                topic_type=self.sub_task_planner_topic_type,
                agent_system=self,
                explorer_topic_type=self.explorer_topic_type,
                solver_topic_type=self.sub_task_solver_topic_type,
                tester_topic_type=self.tester_topic_type,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.sub_task_planner_topic_type, agent_type=self._sub_task_planner_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._sub_task_planner_agent.type))

        self._sub_task_solver_agent = await SubTaskSolverAgent.register(
            self.runtime,
            self.sub_task_solver_topic_type,
            lambda: SubTaskSolverAgent(
                topic_type=self.sub_task_solver_topic_type,
                agent_system=self,
                planner_agent_topic_type=self.sub_task_planner_topic_type,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.sub_task_solver_topic_type, agent_type=self._sub_task_solver_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._sub_task_solver_agent.type))

        self._tester_agent = await TesterAgent.register(
            self.runtime,
            self.tester_topic_type,
            lambda: TesterAgent(
                topic_type=self.tester_topic_type,
                agent_system=self,
                previous_agent_topic_type=self.refiner_topic_type,
                next_agent_topic_type=self.refiner_topic_type,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.tester_topic_type, agent_type=self._tester_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._tester_agent.type))

        self._refiner_agent = await RefinerAgent.register(
            self.runtime,
            self.refiner_topic_type,
            lambda: RefinerAgent(
                topic_type=self.refiner_topic_type,
                agent_system=self,
                tester_topic_type=self.tester_topic_type,
                approver_topic_type=self.approver_topic_type,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.refiner_topic_type, agent_type=self._refiner_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._refiner_agent.type))

        self._approver_agent = await ApproverAgent.register(
            self.runtime,
            self.approver_topic_type,
            lambda: ApproverAgent(
                topic_type=self.approver_topic_type,
                agent_system=self,
                solver_agent_topic_type=self.refiner_topic_type,
                next_agent_topic_type=self.summarizer_topic_type,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.approver_topic_type, agent_type=self._approver_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._approver_agent.type))

        self._summarizer_agent = await SummarizerAgent.register(
            self.runtime,
            self.summarizer_topic_type,
            lambda: SummarizerAgent(
                topic_type=self.summarizer_topic_type,
                agent_system=self,
            ),
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.summarizer_topic_type, agent_type=self._summarizer_agent.type))
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=GROUP_CHAT_TOPIC_TYPE, agent_type=self._summarizer_agent.type))

    async def _trigger_agents(self, user_input: str) -> None:
        """Triggers the execution flow of the agent system's single agents, given the latest user input."""

        await self.runtime.publish_message(
            message=GroupChatMessage(nonce=str(uuid.uuid4()), body=UserMessage(content=user_input, source="user")),
            topic_id=DefaultTopicId(type=GROUP_CHAT_TOPIC_TYPE)
        )
        # 0.1 delay to ensure the GroupChatMessage can be observed before the RequestToSpeak
        # (I think it is not required, however, it certainly does not hurt)
        await asyncio.sleep(0.1)
        await self.runtime.publish_message(
            message=RequestToSpeak(nonce=str(uuid.uuid4())),
            topic_id=DefaultTopicId(type=self.explorer_topic_type)
        )

    def get_state(self):
        """Returns the current state of the agent system for logging and later analysis."""

        return {
            'explorer_agent': self._explorer_agent,
            'sub_task_planner_agent': self._sub_task_planner_agent,
            'sub_task_solver_agent': self._sub_task_solver_agent,
            'tester_agent': self._tester_agent,
            'refiner_agent': self._refiner_agent,
            'approver_agent': self._approver_agent,
            'summarizer_agent': self._summarizer_agent,
        }



import argparse
import asyncio
from pathlib import Path

import nest_asyncio
from dotenv import load_dotenv

from spoox.environment import LocalEnvironment
from spoox.environment.local_environment import ConfirmationMode
from spoox.interface import LogInterface
from spoox.environment.model_clients.utils import setup_model_client, ModelClientId
from spoox.utils import setup_agent_system, AgentSystemId

nest_asyncio.apply()

"""
example usage:
python src/spoox/spoox_headless.py -c openai -m gpt-5-mini -a spoox-m -t "create an empty file named dodo in the current dir"
"""


def main() -> None:
    """
    This CLI script configures and runs a **headless** agent system.
    Executes the agent system with a given task description and exits after the first run.
    Does not wait for interactive user input.
    Typically used for running benchmarks automatically.
    """

    parser = argparse.ArgumentParser(description="Spoox argument parser")
    parser.add_argument(
        "-c",
        "--model-client-id",
        required=True,
        help="Model client, options: 'ollama', 'openai', 'anthropic' (str).",
    )
    parser.add_argument(
        "-m",
        "--model-id",
        required=True,
        help="Model id (str).",
    )
    parser.add_argument(
        "-a",
        "--agent-id",
        required=True,
        help="Agent id (e.g. 'singleton', 'spoox-m') (str)."
    )
    parser.add_argument(
        "-s",
        "--confirmation-mode",
        required=True,
        help="The level of confirmation the agent will seek from the user before interacting with the environment, "
             "options: 'strict', 'self_evaluation', 'no_confirmation' (str).",
    )
    parser.add_argument(
        "-t",
        "--task",
        required=True,
        help="Task description provided to the agent system at start (str).",
    )
    parser.add_argument(
        "-r",
        "--print-reasoning",
        required=False,
        default=True,
        help="Print solution process in terminal (bool).",
    )
    parser.add_argument(
        "-l",
        "--logs-dir",
        required=False,
        default="/tmp/spoox",
        help="Logs dir path (str).",
    )
    parser.add_argument(
        "-x",
        "--max-timeout",
        required=False,
        default=3600,  # 60min default max
        help="Max timeout in seconds (int).",
    )

    args = parser.parse_args()
    client_id = ModelClientId(args.model_client_id)
    model_id = str(args.model_id)
    agent_id = AgentSystemId(args.agent_id)
    print_reasoning = str(args.print_reasoning).lower() in ("yes", "true", "t", "y")
    task = str(args.task)
    logs_dir = Path(str(args.logs_dir))
    max_timeout = int(args.max_timeout)
    confirmation_mode = ConfirmationMode(args.confirmation_mode)

    load_dotenv()

    # setup headless interface -> using log interface
    interface = LogInterface(logging_active=True, print_live=print_reasoning)
    interface.user_delegate.user_input = [task, 'q']
    interface.user_delegate.default_user_choice = 'confirm'

    # setup model client and environment
    model_client = setup_model_client(client_id, model_id)
    environment = LocalEnvironment(interface=interface, confirmation_mode=confirmation_mode)

    # setup and run agent system
    agent = setup_agent_system(agent_id, model_client, environment, interface, max_timeout, logs_dir)
    try:
        asyncio.run(agent.start())
    except Exception as e:
        interface.print_highlight(str(e), f"Exception during agent system execution.")


if __name__ == "__main__":
    main()



from enum import Enum
from pathlib import Path
from typing import Optional

from autogen_core import BaseAgent, FunctionCall
from autogen_core.tools import BaseTool
from autogen_ext.tools.langchain import LangChainToolAdapter
from langchain_community.tools import DuckDuckGoSearchResults

from spoox.environment import Environment
from spoox.environment.code_executors import CodeExecutorLocal
from spoox.environment.tools import ShellTool, PythonTool
from spoox.environment.tools import TerminalTool
from spoox.interface import Interface


class ConfirmationMode(Enum):
    """The level of confirmation the agent will seek from the user before interacting with the environment."""

    STRICT = 'strict'  # Agent seeks user confirmation before each environment interaction.
    SELF_EVALUATION = 'self_evaluation'  # Agent autonomously decides when confirmation is needed (e.g., `cd` does not require confirmation, while `rm` commands always do).
    NO_CONFIRMATION = 'no_confirmation'  # Agent never seeks user confirmation. Warning: Use only in isolated or sandboxed environments.


class LocalEnvironment(Environment):
    """
    This is a concrete environment implementation providing basic programming tools and access the local computer.
    Also, `get_additional_tool_descriptions` and `get_tools` are configured for each relevant specific spoox agent.
    """

    def __init__(self, interface: Interface, confirmation_mode: ConfirmationMode,
                 work_dir: Optional[Path] = None, user: Optional[str] = None):
        super().__init__(interface)

        self.confirmation_mode = confirmation_mode
        self._code_executor = CodeExecutorLocal(work_dir=work_dir, user=user)

        self._shell_tool = ShellTool(self._code_executor)
        self._terminal_tool = TerminalTool()
        self._python_tool = PythonTool(self._code_executor)
        self._search_tool = LangChainToolAdapter(DuckDuckGoSearchResults(output_format="list"))

    async def start(self):
        """Stop the environment. Should be called once during agent system shutdown."""
        await self._code_executor.start()
        self._started = True

    async def stop(self):
        """Stop the environment. Should be called once during agent system shutdown."""
        await self._code_executor.stop()
        await self._terminal_tool.stop()
        self._started = False

    async def reset(self):
        """
        Resets the environment. Typically called when an agent starts working in a shared environment
        to ensure no dependencies exist from previous agent operations.
        """
        await self._code_executor.restart()
        await self._terminal_tool.reset()

    def get_tools(self, agent: BaseAgent) -> list[BaseTool]:
        """Returns a list of tools the agent should be equipped with. All relevant spoox agents are represented."""

        class_name = agent.__class__.__name__
        if class_name in ["SingletonAgent"]:
            return [self._shell_tool, self._python_tool, self._terminal_tool, self._search_tool]
        elif class_name in ["ExplorerAgent"]:
            return [self._shell_tool, self._search_tool, self._terminal_tool]
        elif class_name in ["SolverAgent", "SubTaskSolverAgent", "RefinerAgent"]:
            return [self._shell_tool, self._python_tool, self._terminal_tool, self._search_tool]
        elif class_name in ["TesterAgent", "ApproverAgent"]:
            return [self._shell_tool, self._python_tool, self._terminal_tool]
        # default: return no tools (e.g. Approver, Summarizer, ...)
        return []

    def get_additional_tool_descriptions(self, agent: BaseAgent) -> [str]:
        """
        Agent system prompts may include additional descriptions for their environment and tools.
        All relevant spoox agents are represented.
        """

        class_name = agent.__class__.__name__
        shell_descr = f"""- Use Shell tool for simple, single bash commands. Do **not** use it for commands that open interactive terminal programs (e.g. git log or man)."""
        py_descr = f"""- Use PythonExecutor tool for complex logic, scripting, or data processing."""
        terminal_descr = f"""- Use Terminal tool to execute commands in a persistent terminal instance and get back the exact terminal screen. This is especially useful for interactive programs such as git log, man, or vim."""

        if class_name in ["SingletonAgent"]:
            return [shell_descr, py_descr, terminal_descr]
        elif class_name in ["ExplorerAgent"]:
            return [shell_descr, terminal_descr]
        elif class_name in ["SolverAgent", "SubTaskSolverAgent", "RefinerAgent"]:
            return [shell_descr, py_descr, terminal_descr]
        elif class_name in ["TesterAgent", "ApproverAgent"]:
            return [shell_descr, py_descr, terminal_descr]
        # default: return no additional tool descriptions (e.g. Approver, Summarizer, ...)
        return []

    def _check_tool_call_confirmation(self, call: FunctionCall) -> str:
        """
        Check if user configuration should be seeked for the given FunctionCall.
        Returns an empty string if call is confirmed and can be executed, or a rejection message.
        For the LocalEnvironment this is dependent on the selected confirmation_mode.
        """
        restricted_tool_names = ["shell", "python", "terminal", "codeexecutor"]
        confirmation_choices = ['confirm', 'reject']

        # NO confirmation required
        if call.name.lower() not in restricted_tool_names:
            return ""
        if self.confirmation_mode == ConfirmationMode.NO_CONFIRMATION:
            return ""

        # confirmation required
        user_choice = 'reject'
        if self.confirmation_mode == ConfirmationMode.STRICT:
            user_choice = self.interface.request_select_choice(
                f"Please confirm whether to proceed with the latest {call.name} tool call:",
                confirmation_choices
            )
        elif self.confirmation_mode == ConfirmationMode.SELF_EVALUATION:
            if self._risk_in_code_execution(call.arguments):
                user_choice = self.interface.request_select_choice(
                    f"Please confirm whether to proceed with the latest {call.name} tool call:",
                    confirmation_choices
                )
            else:
                user_choice = 'confirm'
        if user_choice == 'reject':
            return self.interface.request_user_input(
                "Please explain why you chose to reject this tool call:",
                default="Command poses a potential security risk."
            )
        return ""

    def _risk_in_code_execution(self, code: str) -> bool:
        """
        Risk assessment if code snippet can be executed without user confirmation.
        Returns true if risk detected and user confirmation is required.
        """
        # todo better LLM implementation (already added to git issues)
        if any(r in code for r in self.RISK_KEYWORDS):
            return True
        return False

    RISK_KEYWORDS = [
        # --- File system / destructive ---
        "rm ", "rm -", "rm -rf", "unlink", "shutil.rmtree",
        "os.remove", "os.unlink", "os.rmdir",
        "shred", "dd if=", "mkfs", "wipefs",
        # --- Process / execution ---
        "exec(", "eval(", "compile(",
        "subprocess.", "os.system", "os.popen",
        "Popen(", "call(", "run(",
        "bash -c", "sh -c",
        # --- Privilege escalation ---
        "sudo", "su ", "setuid", "setgid",
        # --- Networking / remote code ---
        "curl ", "wget ", "ftp ", "scp ", "rsync ",
        "requests.get", "requests.post",
        "socket.", "paramiko",
        # --- Persistence / system modification ---
        "crontab", "/etc/passwd", "/etc/shadow",
        "systemctl", "service ",
        "chkconfig",
        "~/.bashrc", "~/.profile",
        # --- Disk / hardware ---
        "/dev/sd", "/dev/nvme", "mount ", "umount ",
        # --- Containers / virtualization ---
        "docker run", "docker exec", "docker rm",
        "kubectl", "helm ",
        # --- Environment manipulation ---
        "export ", "unset ",
        "os.environ",
        # --- Potential fork bombs / abuse ---
        ":(){", "fork(", "multiprocessing"
    ]



from .agent_system_spoox_large import SpooxLarge
from .agent_system_spoox_medium import SpooxMedium
from .agent_system_spoox_small import SpooxSmall
from .base_agent import BaseGroupChatAgent


__all__ = [
    "BaseGroupChatAgent",
    "SpooxLarge",
    "SpooxMedium",
    "SpooxSmall",
]


from .cli_interface import CLInterface
from .log_interface import LogInterface
from .interface import Interface
from .interface import CLIColor


__all__ = [
    "CLInterface",
    "LogInterface",
    "Interface",
    "CLIColor",
]



from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.base_agent import BaseGroupChatAgent
from spoox.agents.mas.agents.prompts import get_SUB_TASK_PLANNER_SYSTEM_MESSAGE


class SubTaskPlannerAgent(BaseGroupChatAgent):
    """
    Motivated by the idea of incorporating task decomposition into the MAS,
    the Sub-Task Planner identifies and formulates the next appropriate sub-task given the system’s current progress.
    Its role is to describe the sub-task and sketch a high-level approach for solving it,
    expressed purely in natural language, without offering a detailed step-by-step plan.
    To reinforce this abstraction-focused behavior, the agent is intentionally not equipped with any tools.
    """

    def __init__(
            self,
            topic_type: str,
            agent_system: AgentSystem,
            explorer_topic_type: str,
            solver_topic_type: str,
            tester_topic_type: str,
    ) -> None:

        system_message = get_SUB_TASK_PLANNER_SYSTEM_MESSAGE(
            topic_type, explorer_topic_type, solver_topic_type, tester_topic_type)

        super().__init__(
            description="Agent tasked to create a plan for solving the task or a sub-task.",
            system_message=system_message,
            agent_system=agent_system,
            next_agent_topic_types=[explorer_topic_type, solver_topic_type, tester_topic_type],
            max_internal_iterations=10,
        )



from pathlib import Path

from autogen_core import DefaultTopicId
from autogen_core import TypeSubscription
from autogen_core.models import UserMessage, ChatCompletionClient

from spoox.agents.agent_system import AgentSystem
from spoox.agents.singleton.messages import PublicMessage
from spoox.agents.singleton.singelton_agent import SingletonAgent
from spoox.environment import Environment
from spoox.environment.model_clients.custom_client_openai import CustomOpenAIResponseAPIClient
from spoox.interface import Interface


class SingletonAgentSystem(AgentSystem):
    """
    This is the simplest agent system, consisting of a single agent.
    After the user submits a prompt, the singleton agent is executed, performs its work
    across multiple internal iterations, and terminates upon completion.
    """

    singleton_topic_type = "singleton"

    def __init__(self, interface: Interface, model_client: ChatCompletionClient | CustomOpenAIResponseAPIClient,
                 environment: Environment, timeout: int = 600, logs_dir: Path = Path.cwd()):

        super().__init__(interface, model_client, environment, timeout, logs_dir)
        self._singleton_agent = None

    async def _build_agents(self):
        """Initializing all agents, including all message subscriptions."""

        self._singleton_agent = await SingletonAgent.register(
            self.runtime,
            self.singleton_topic_type,
            lambda: SingletonAgent(self)
        )
        await self.runtime.add_subscription(
            TypeSubscription(topic_type=self.singleton_topic_type, agent_type=self._singleton_agent.type))

    async def _trigger_agents(self, user_input: str) -> None:
        """Triggers the execution flow of the agent system's single agents, given the latest user input."""

        await self.runtime.publish_message(
            message=PublicMessage(body=UserMessage(content=user_input, source="user")),
            topic_id=DefaultTopicId(type=self.singleton_topic_type),
        )

    def get_state(self):
        """Returns the current state of the agent system for logging and later analysis."""
        return {'single_agent_type': self._singleton_agent}



import os
import subprocess
import tempfile
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, ClassVar

from autogen_core import CancellationToken, Component
from autogen_core.code_executor import CodeExecutor, CodeBlock, CodeResult
from pydantic import BaseModel


@dataclass
class CodeBlockTimeout(CodeBlock):
    """A code block extracted from an agent message. Including max timeout."""
    timeout: Optional[int] = None


class CodeExecutorLocalConfig(BaseModel):
    """Configuration for PermanentCommandLineCodeExecutor"""
    timeout_min: Optional[int] = 10
    timeout_max: Optional[int] = 90
    work_dir: Optional[str] = None


class CodeExecutorLocal(CodeExecutor, Component[CodeExecutorLocalConfig]):
    """
    This CodeExecutor executes Python and Bash code blocks locally by writing them to temporary files,
    running them as subprocesses with configurable timeouts and working directories,
    and returning the execution results with stdout/stderr output.
    """

    component_config_schema = CodeExecutorLocalConfig

    SUPPORTED_LANGUAGES: ClassVar[List[str]] = [
        "python",
        "bash",
    ]

    def __init__(self, timeout_min: Optional[int] = 10, timeout_max: Optional[int] = 90,
                 work_dir: Optional[Path] = None, user: Optional[str] = None):
        super().__init__()

        if timeout_min < 1 or timeout_max < 1:
            raise ValueError("Timeout values must be greater than or equal to 1.")
        self.timeout_min = timeout_min
        self.timeout_max = timeout_max

        self._started = False
        self._runs_count = 0
        self._local_code_dir = None
        self._local_code_dir_path = None

        self.work_dir = work_dir or Path.cwd()
        self.user = user or os.environ.get("USER") or os.environ.get("USERNAME")

    @property
    def runs_count(self) -> int:
        """Counter of executed code blocks."""
        return self._runs_count

    async def execute_code_blocks(self, code_blocks: List[CodeBlock],
                                  cancellation_token: CancellationToken) -> CodeResult:
        """
        This method validates and executes a single code block by writing it to a temporary file,
        running it as a subprocess with a bounded timeout, and returning the execution result
        with formatted stdout/stderr output or a timeout error.
        """

        if len(code_blocks) != 1:
            raise RuntimeError(f"CodeExecutorLocal `code_blocks` must exactly contain one code block.")
        code_block = code_blocks[0]

        if not self._started:
            raise RuntimeError(f"CodeExecutorLocal must be started. Make sure `.start()` is called.")

        if code_block.language.lower() not in self.SUPPORTED_LANGUAGES:
            raise ValueError(f"Language '{code_block.language}' not supported by CodeExecutorLocal.")

        # setup code block language dependents
        type_of_file = ".py" if code_block.language.lower() == "python" else ".sh"
        caller_keyword = "python" if code_block.language.lower() == "python" else "bash"

        # write code block to the py or sh file
        rand_id = uuid.uuid4().hex[:8]
        code_block_file = self._local_code_dir_path / f"code_block_{self._runs_count}_{rand_id}{type_of_file}"
        code_block_file.write_text(f"{code_block.code}\n")
        code_block_file.chmod(0o755)

        # extract timeout if available and ensure it stays within defined min max boundaries
        if isinstance(code_block, CodeBlockTimeout) and code_block.timeout is not None:
            timeout = code_block.timeout
            if timeout < self.timeout_min:
                timeout = self.timeout_min
            elif timeout > self.timeout_max:
                timeout = self.timeout_max
        else:
            timeout = self.timeout_min

        # exec code file
        try:
            result = subprocess.run(
                [caller_keyword, str(code_block_file)],
                capture_output=True,
                text=True,
                timeout=timeout,
                user=self.user,
                cwd=str(self.work_dir)
            )
        except subprocess.TimeoutExpired:
            return CodeResult(1, f"Command execution failed - timeout of {timeout}s exceeded")

        # format exec output
        if result.stdout and result.stderr:
            return CodeResult(result.returncode,
                              f"<stdout>\n{result.stdout}\n</stdout>\n\n<stderr>\n{result.stderr}\n</stderr>")
        elif result.stdout:
            return CodeResult(result.returncode, result.stdout)
        elif result.stderr:
            return CodeResult(result.returncode, result.stderr)
        return CodeResult(result.returncode, "")

    async def start(self) -> None:
        if not self._started:
            # setup local temp code dir
            self._local_code_dir = tempfile.TemporaryDirectory(dir=Path("/tmp"))
            self._local_code_dir_path = Path(self._local_code_dir.name)
            # others
            self._runs_count = 0
            self._started = True

    async def stop(self) -> None:
        if self._started:
            self._local_code_dir.cleanup()
            self._started = False

    async def restart(self) -> None:
        if self._started:
            await self.stop()
        await self.start()



from typing import Optional

from autogen_core import CancellationToken, Component, ComponentModel
from autogen_core.code_executor import CodeExecutor
from autogen_core.tools import BaseTool
from pydantic import BaseModel, Field, model_serializer
from typing_extensions import Self

from spoox.environment.code_executors.code_executor_local import CodeBlockTimeout
from spoox.environment.tools.utils import output_truncat

"""
For human developers, opening a terminal and executing simple Bash commands serves as the standard lightweight interface 
for performing general server management and diverse OS tasks. The action space is almost unbounded, multiple commands 
can be composed, modified, and chained, to accomplish complex objectives. The Shell tool shall make this powerful 
functionality available to agents. It has intentionally been designed with minimal complexity. To invoke it, the agent 
only needs to provide the command as a string. The command is always executed in the current user’s home directory, 
and the resulting shell output, including the exit code, is stored to the agent’s context. Thereby, the execution 
environment is not persistent. Commands that manipulate the shell session, such as cd, do not affect future tool calls.
"""


class CodeExecutionInput(BaseModel):
    """Input object for ShellTool."""

    command: str = Field(description="The Bash command that should be executed in the shell.")
    timeout: int = Field(
       description="(Optional) Maximum duration (in seconds) the command may run. "
                   "Define this only if you need to increase the default timeout of 10s. "
                   "Timeout value must not exceed 90s.",
       default=10
    )


class CodeExecutionResult(BaseModel):
    """Result object for ShellTool."""

    output: str
    exit_code: Optional[int] = None
    truncated: bool = False

    @model_serializer
    def ser_model(self) -> str:
        out = ""
        if self.exit_code is not None:
            out += f"<exit-code> {self.exit_code} </exit-code> \n"
        if self.truncated:
            out += f"<truncated-output> True </truncated-output> \n"
        out += f"<output> {self.output} </output>"
        return out


class ShellToolConfig(BaseModel):
    """Configuration for ShellTool."""

    executor: ComponentModel
    output_max: int
    description: str = "Execute a Bash command in the shell, in the users current directory."


class ShellTool(BaseTool[CodeExecutionInput, CodeExecutionResult], Component[ShellToolConfig]):
    """A tool that executes Bash code in a code executor and returns the output."""

    component_config_schema = ShellToolConfig

    def __init__(self, executor: CodeExecutor, output_max: int = 10000):
        super().__init__(
            CodeExecutionInput,
            CodeExecutionResult,
            "Shell",
            "Execute a Bash command in the shell, in the users current directory.",
            strict=True
        )
        self._executor = executor
        self._output_max = output_max

    async def run(self, args: CodeExecutionInput, cancellation_token: CancellationToken) -> CodeExecutionResult:
        # execute code
        code_block = CodeBlockTimeout(code=args.command, language="bash", timeout=args.timeout)
        result = await self._executor.execute_code_blocks(
            code_blocks=[code_block], cancellation_token=cancellation_token
        )
        # make sure the output is cut when too long
        output, truncated = output_truncat(result.output, self._output_max)
        return CodeExecutionResult(exit_code=result.exit_code, output=output, truncated=truncated)

    def _to_config(self) -> ShellToolConfig:
        """Convert current instance to config object."""
        return ShellToolConfig(executor=self._executor.dump_component(), output_max=self._output_max)

    @classmethod
    def _from_config(cls, config: ShellToolConfig) -> Self:
        """Create instance from config object."""
        executor = CodeExecutor.load_component(config.executor)
        return cls(executor=executor, output_max=config.output_max)



import asyncio
import json
import pickle
import time
from abc import abstractmethod, ABC
from datetime import datetime
from pathlib import Path

from autogen_core import SingleThreadedAgentRuntime
from autogen_core.models import ChatCompletionClient
from spoox.environment import Environment
from spoox.environment.model_clients.custom_client_openai import CustomOpenAIResponseAPIClient
from spoox.interface import Interface


class AgentSystem(ABC):
    """
    This class serves as the base for all agent system implementations.
    It provides fundamental runtime functionality such as the human-agent interaction loop (see `start()`),
    which requests user input and executes agents in response.
    Additionally, it implements timeout management and comprehensive logging support.

    Concrete agent system implementations just have to implement:

    - `_build_agents()` that initializes all concreate single agents and their autogen message subscriptions.
    - `_trigger_agents()` implementing the triggering of the agent flow given the latest user intput message, typically realized by publishing an autogen message.
    - `get_state()` simple function returning the most important objects within a dictionary that should be logged and could be useful for post-analysis.
    """

    def __init__(self, interface: Interface, model_client: ChatCompletionClient | CustomOpenAIResponseAPIClient,
                 environment: Environment, timeout: int = 600, logs_dir: Path = Path.cwd()):

        self.interface = interface
        self.model_client = model_client
        self.environment = environment
        self.timeout = timeout

        self.runtime = SingleThreadedAgentRuntime()

        timestamp = datetime.now().strftime("%Y-%m-%d__%H-%M-%S")
        self.logs_dir = logs_dir / f"spoox_logs_{timestamp}"
        self.logs_dir.mkdir(parents=True, exist_ok=True)
        self.usage_stats = dict()
        self.init_usage_stats()

        # event passed to single agents to notify them when a timeout occurs
        self._timeout_event = asyncio.Event()
        # async function that contains the timeout countdown if started
        self._timeout_countdown = None

    @property
    def timeout_event(self) -> asyncio.Event:
        return self._timeout_event

    async def start(self) -> None:
        """
        Start and run the agent system.
        The agent system is initialized and enters an infinite loop that alternates between
        waiting for user input and executing the agents by calling self._trigger_agents.
        The system exits when the user enters 'q', 'exit', or 'stop'.
        """

        # start the agent system
        await self.environment.start()
        await self._build_agents()
        self.save_logs()
        start_time = time.time()

        # user input and agent calling loop
        while True:

            user_input = self.interface.request_user_input("Query...")
            if user_input in ['q', 'exit', 'stop']:
                break

            self.runtime.start()
            await self._trigger_agents(user_input)
            self._start_timeout_countdown()
            await self.runtime.stop_when_idle()
            self._cancel_timeout_countdown()
            self.save_logs()

        # stop entirely
        await self.environment.stop()
        await self.runtime.close()
        self.save_logs(stopped=True, exec_time_sec=int(time.time() - start_time))

    def init_usage_stats(self) -> None:
        """
        Initializes a shared dictionary for collecting interesting statistics.
        This dictionary is provided to all single agents to enable a centralized collection.
        """

        self.usage_stats['llm_calls_count'] = 0
        self.usage_stats['tool_call_counts'] = dict()
        self.usage_stats['tool_calls'] = []
        self.usage_stats['model_client_exceptions'] = []
        self.usage_stats['agent_errors'] = []
        self.usage_stats['prompt_tokens'] = []
        self.usage_stats['completion_tokens'] = []
        self.usage_stats['next_agent_calling_chain'] = []
        self.usage_stats['group_chat_message_lengths'] = []

    def _cancel_timeout_countdown(self) -> None:
        """Cancels the timeout event without triggering or restarting the timeout countdown."""

        if self._timeout_countdown is not None and not self._timeout_countdown.done():
            self._timeout_countdown.cancel()
        self._timeout_countdown = None

        if self._timeout_event.is_set():
            self._timeout_event.clear()

    def _start_timeout_countdown(self) -> None:
        """
        Starts a timeout countdown for self.timeout seconds. Once the timeout is reached, self.timeout_event is set.
        The self.timeout_event is shared with all individual agents, allowing them to terminate their execution
        gracefully once the event is set. Directly stopping the runtime while agents are still running
        can result in a ValueError being raised by the autogen runtime. To avoid this, an event is used to notify
        all agents to stop as soon as possible instead of abruptly terminating execution.
        """

        async def _timeout():
            await asyncio.sleep(self.timeout)
            error_message = f"Agent system timeout after {self.timeout}s."
            self.interface.print_highlight(error_message, "TimeoutError")
            self.usage_stats["agent_errors"].append(("TimeoutError", error_message))
            self._timeout_event.set()

        self._cancel_timeout_countdown()
        self._timeout_countdown = asyncio.create_task(_timeout())

    def save_logs(self, stopped: bool = False, exec_time_sec: int = 0) -> None:
        """
        Store agent system logs, including usage_stats, get_state dictionary, the entire interface and some meta-data.
        Execution time is under 1 ms, making it suitable for frequent use during agent operation.
        """

        with (self.logs_dir / f"meta_data.json").open("w") as f:
            meta_data = {
                "agent-system-type": self.environment.__class__.__name__,
                "model-info": self.model_client.model_info,
                "environment-type": self.environment.__class__.__name__,
                "interface-type": self.interface.__class__.__name__,
                "timeout": self.timeout,
                "model-client-actual-usage-prompt-tokens": self.model_client.actual_usage().prompt_tokens,
                "model-client-actual-usage-completion-tokens": self.model_client.actual_usage().completion_tokens,
                "model-client-total-usage-prompt-tokens": self.model_client.total_usage().prompt_tokens,
                "model-client-total-usage-completion-tokens": self.model_client.total_usage().completion_tokens,
                "agent-system-stopped": stopped,
                "agent-system-exec-time-sec": exec_time_sec,
            }
            json.dump(meta_data, f, indent=4)
        # save the agent_usage_stats dict in a pickle file
        with (self.logs_dir / f"usage_stats.pkl").open("wb") as f:
            pickle.dump(self.usage_stats, f)
        # save the agent system state as a dict in a pickle file
        with (self.logs_dir / f"agent_state.pkl").open("wb") as f:
            pickle.dump(self.get_state(), f)
        # save the interface logs in a pickle file
        with (self.logs_dir / f"interface.pkl").open("wb") as f:
            pickle.dump(self.interface, f)

    @abstractmethod
    async def _build_agents(self) -> None:
        """
        Initializing all agents, including all message subscriptions.
        Must be implemented by the respective agent system implementation.
        """
        pass

    @abstractmethod
    async def _trigger_agents(self, user_input: str) -> None:
        """
        Triggers the execution flow of the agent system's single agents, given the latest user input.
        Must be implemented by the respective agent system implementation.
        """
        pass

    @abstractmethod
    def get_state(self) -> dict:
        """
        Returns the current state of the agent system for logging and later analysis.
        Must be implemented by the respective agent system implementation.
        """
        pass





# basic prompt blocks used by most agents


MORE_OPERATING_RULES = [
    f"""""",
    f"""# More Operating Rules""",
    f"""- Think and act step by step and explain your actions and reasoning. """,
    f"""- Break complex objectives into smaller sub-goals if necessary.""",
    f"""- Use tools to gather knowledge or verify assumptions before proceeding.""",
    f"""- If required context is missing, do not guess, use available tools to retrieve it first.""",
    f"""- If you call a tool it will be executed and results will be returned to you."""
]


def get_CODE_EXECUTION_CAPABILITIES(additional_tool_descriptions: [str]) -> [str]:
    _GENERAL_CODE_INSTRUCTIONS = [
        f"""""",
        f"""# Code Execution Capabilities""",
        f"""- Your are equipped with different tools for executing code.""",
        f"""- **Do only use the tools to execute code.** Code inside a Markdown Code Block will **not** be executed.""",
    ]
    return _GENERAL_CODE_INSTRUCTIONS + additional_tool_descriptions



from spoox.agents.agent_system import AgentSystem
from spoox.agents.mas.base_agent import BaseGroupChatAgent
from spoox.agents.mas.agents.prompts import get_SUMMARIZER_SYSTEM_MESSAGE


class SummarizerAgent(BaseGroupChatAgent):
    """
    This agent has been introduced mainly for usability purposes, since clear, user-facing explanations contribute
    to more trustworthy AI4SE systems. Its only role is to craft a concise and user-oriented final summary
    based on the group chat history, without using any tools.
    """

    def __init__(
            self,
            topic_type: str,
            agent_system: AgentSystem,
    ) -> None:

        super().__init__(
            description="Agent creating the final summary.",
            system_message=get_SUMMARIZER_SYSTEM_MESSAGE(topic_type),
            agent_system=agent_system,
            max_internal_iterations=10,
            skip=True  # todo remove
        )
